# SPDX-License-Identifier: GPL-2.0

import datetime
import signal
import subprocess

import _damo_fmt_str
import _damo_subproc
import _damo_sysinfo
import _damon

tracer_pipe = None
output_file = None

def signalhandler(signum, frame):
    print('signal %s received' % signum)
    if tracer_pipe:
        try:
            tracer_pipe.send_signal(signum)
            tracer_pipe.wait()
        except:
            pass
    if output_file:
        output_file.close()
    exit(0)

def read_perf_record(record_file):
    try:
        trace_text = subprocess.check_output(
                ['perf', 'script', '--force', '-i', record_file]).decode()
    except Exception as e:
        return None, '%s' % e
    return trace_text, None

def read_trace_cmd_record(record_file):
    try:
        trace_text = subprocess.check_output(
                ['trace-cmd', 'report', '-i', record_file]).decode()
    except Exception as e:
        return None, '%s' % e
    return trace_text, None

def read_damo_report_trace_output(output_file):
    try:
        with open(output_file, 'r') as f:
            return f.read(), None
    except Exception as e:
        return None, '%s' % e

def read_trace_record(record_file):
    trace_text, perf_err = read_perf_record(record_file)
    if perf_err is None:
        return trace_text, 'perf-script', None
    trace_text, trace_cmd_err = read_trace_cmd_record(record_file)
    if trace_cmd_err is None:
        return trace_text, 'trace-cmd-report', None
    trace_text, text_err = read_damo_report_trace_output(record_file)
    if text_err is None:
        tracer = trace_text.split('\n')[2].split()[1]
        return trace_text, 'damo-report-trace-%s' % tracer, None
    err = 'cannot parse %s via perf (%s), trace-cmd (%s), file read (%s)' % (
            record_file, perf_err, trace_cmd_err, text_err)
    return None, None, err

damon_trace_events = _damo_sysinfo.tracepoint_to_feature_name_map.keys()

def get_events_to_show(to_show, to_hide):
    events = []

    if 'all' in to_show:
        events = list(damon_trace_events)
    else:
        events = to_show
    if to_hide is not None:
        for event in to_hide:
            events.remove(event)
    return events

def pr_wrapped(line, max_cols):
    if len(line) <= max_cols:
        print(line)
        return
    fields = line.split()
    line_fields = []
    len_line = 0
    for field in fields:
        len_field = len(field)
        if len_line + len_field > max_cols:
            if len(line_fields) > 0:
                print(' '.join(line_fields))
                line_fields = []
            else:
                print(field)
            # indent second and later lines
            line_fields.append('  ')
            len_line = 2
        line_fields.append(field)
        len_line += len_field
        if len(line_fields) > 1:
            len_line += 1
    if len(line_fields) > 0:
        print(' '.join(line_fields))

def get_trace_fields(fields, trace_text_format, trace_name):
    if trace_text_format == 'damo-report-trace-perf':
        #   3128.371 kdamond.0/764 damon:damon_aggregated(trace fields)
        trace_fields = fields[2:]
        trace_fields[0] = trace_fields[0][len(trace_name) + 1:]
        trace_fields[-1] = trace_fields[-1][:-1]
    else:
        # <...>-764   [001] .....  1394.412830: damon_region_aggregated: trace fields
        trace_fields = fields[3:]
    return trace_fields

region_idx = 0
def pr_damon_aggregated(fields, trace_text_format, max_cols):
    trace_fields = get_trace_fields(
            fields, trace_text_format, 'damon:damon_aggregated')
    # trace_fields: target_id=0 nr_regions=11 8009068544-8372879360: 0 2740
    target_id = int(trace_fields[0].split('=')[1])
    nr_regions = int(trace_fields[1].split('=')[1])
    start = int(trace_fields[2].split('-')[0])
    end = int(trace_fields[2].split('-')[1][:-1])
    nr_accesses = int(trace_fields[3])
    if len(trace_fields) > 4:
        age = int(trace_fields[4])
    else:
        age = -1

    global region_idx
    trace_text = '%d %d/%d %s (%s) %d %d' % (
            target_id, region_idx, nr_regions,
            _damo_fmt_str.format_sz_accurate(start, machine_friendly=False),
            _damo_fmt_str.format_sz_accurate(
                end - start, machine_friendly=False),
            nr_accesses, age)
    region_idx += 1
    if region_idx == nr_regions:
        region_idx = 0
    pr_wrapped(' '.join(fields[:2] + ['damon_aggregated', trace_text]),
               max_cols)

region_aggregated_idx = 0
def fmt_damon_region_aggregated_trace(trace_fields):
    # trace_fields: target_id=0 nr_regions=11 8009068544-8372879360: 0 2740 probe_hits=14 00
    target_id = int(trace_fields[0].split('=')[1])
    nr_regions = int(trace_fields[1].split('=')[1])
    start = int(trace_fields[2].split('-')[0])
    end = int(trace_fields[2].split('-')[1][:-1])
    nr_accesses = int(trace_fields[3])
    age = int(trace_fields[4])
    if trace_fields[5] == 'probe_hits=':
        probe_hits = []
    else:
        probe_hits = [int(hits, 16) for hits in
                      [trace_fields[5].split('=')[1]] + trace_fields[6:]]

    global region_aggregated_idx
    trace_text = '%d %d/%d %s (%s) %d %d %s' % (
            target_id, region_aggregated_idx, nr_regions,
            _damo_fmt_str.format_sz_accurate(start, machine_friendly=False),
            _damo_fmt_str.format_sz_accurate(
                end - start, machine_friendly=False),
            nr_accesses, age, ' '.join(['%d' % h for h in probe_hits]))
    region_aggregated_idx += 1
    if region_aggregated_idx == nr_regions:
        region_aggregated_idx = 0
    return trace_text

def fmt_damon_region_aggregated(fields, trace_text_format):
    trace_fields = get_trace_fields(
            fields, trace_text_format, 'damon:damon_region_aggregated')
    trace_text = fmt_damon_region_aggregated_trace(trace_fields)
    return ' '.join(fields[:2] + ['damon_region_aggregated', trace_text])

def pr_damon_region_aggregated(fields, trace_text_format, max_cols):
    pr_wrapped(
            fmt_damon_region_aggregated(fields, trace_text_format), max_cols)

def pr_damos_before_apply(fields, trace_text_format, max_cols):
    trace_fields = get_trace_fields(
            fields, trace_text_format, 'damon:damos_before_apply')
    # trace_fields: ctx_idx=0 scheme_idx=0 target_idx=0 nr_regions=11 1234-5678: 10 45
    context_idx = int(trace_fields[0].split('=')[1])
    scheme_idx = int(trace_fields[1].split('=')[1])
    target_idx = int(trace_fields[2].split('=')[1])
    nr_regions = int(trace_fields[3].split('=')[1])
    start = int(trace_fields[4].split('-')[0])
    end = int(trace_fields[4].split('-')[1][:-1])
    nr_accesses = int(trace_fields[5])
    age = int(trace_fields[6])

    trace_text = '%d %d %d %d %s (%s) %d %d' % (
            context_idx, scheme_idx, target_idx, nr_regions,
            _damo_fmt_str.format_sz_accurate(start, machine_friendly=False),
            _damo_fmt_str.format_sz_accurate(
                end - trace_data['start'], machine_friendly=False),
            nr_accesses, age)
    pr_wrapped(' '.join(fields[:2] + ['damos_before_apply', trace_text]),
               max_cols)

def pr_damos_stat(fields, trace_text_format, max_cols):
    trace_fields = get_trace_fields(
            fields, trace_text_format, 'damon:damos_stat')
    context_idx = int(trace_fields[0].split('=')[1])
    scheme_idx = int(trace_fields[1].split('=')[1])
    nr_tried = int(trace_fields[2].split('=')[1])
    sz_tried = int(trace_fields[3].split('=')[1])
    nr_applied = int(trace_fields[4].split('=')[1])
    sz_applied = int(trace_fields[5].split('=')[1])
    if len(trace_fields) > 6:
        sz_ops_filter_passed = int(trace_fields[6].split('=')[1])
    else:
        sz_ops_filter_passed = -1
    if len(trace_fields) > 7:
        qt_exceeds = int(trace_fields[7].split('=')[1])
    else:
        qt_exceeds = -1
    if len(trace_fields) > 8:
        nr_snapshots = int(trace_fields[8].split('=')[1])
    else:
        nr_snapshots = -1

    trace_text = '%d %d %s %s %s %s %s %s %s' % (
            context_idx, scheme_idx,
            _damo_fmt_str.format_nr(nr_tried, machine_friendly=False),
            _damo_fmt_str.format_sz_accurate(
                sz_tried, machine_friendly=False),
            _damo_fmt_str.format_nr(nr_applied, machine_friendly=False),
            _damo_fmt_str.format_sz_accurate(
                sz_applied, machine_friendly=False),
            _damo_fmt_str.format_sz_accurate(
                sz_ops_filter_passed, machine_friendly=False),
            _damo_fmt_str.format_nr(qt_exceeds, machine_friendly=False),
             _damo_fmt_str.format_nr(nr_snapshots, machine_friendly=False)
             )
    pr_wrapped(' '.join(fields[:2] + ['damos_stat', trace_text]), max_cols)

def pr_trace_line(line, raw, trace_text_format, max_cols):
    if raw is True:
        print(line)
    fields = line.split()
    if trace_text_format == 'perf-script':
        fields = [fields[1]] + fields[3:]
        event = fields[2]
        event = event[len('damon:'):]
        fields[2] = event

        kdamond_pid = fields[0]
        timestamp = fields[1]
    elif trace_text_format == 'trace-cmd-report':
        fields = [fields[0]] + fields[3:]

        kdamond_pid = fields[0].split('-')[-1]
        timestamp = fields[1]
    elif trace_text_format == 'damo-report-trace-perf':
        fields = fields
        event = fields[2]
        event = event[len('damon:'):]
        fields[2] = event

        kdamond_pid = fields[1].split('/')[-1]
        timestamp = fields[0]
    elif trace_text_format == 'damo-report-trace-trace-cmd':
        fields = [fields[0]] + fields[3:]

        kdamond_pid = fields[0].split('-')[-1]
        timestamp = fields[1]

    fields[0] = timestamp
    fields[1] = kdamond_pid

    if fields[2].startswith('damon_region_aggregated'):
        pr_damon_region_aggregated(fields, trace_text_format, max_cols)
        return
    if fields[2].startswith('damon_aggregated'):
        pr_damon_aggregated(fields, trace_text_format, max_cols)
        return
    if fields[2].startswith('damos_before_apply'):
        pr_damos_before_apply(fields, trace_text_format, max_cols)
        return
    if fields[2].startswith('damos_stat_after_apply_interval'):
        pr_damos_stat(fields, trace_text_format, max_cols)
        return
    pr_wrapped(' '.join(fields), max_cols)

def report_recorded_trace(args):
    trace_text, trace_text_format, err = read_trace_record(args.input)
    if err is not None:
        print(err)
        return -1

    if args.event is None:
        print('--event is required')
        exit(1)

    events = get_events_to_show(args.event, args.no_event)

    for line in trace_text.split('\n'):
        fields = line.split()
        if trace_text_format in ['perf-script', 'trace-cmd-report']:
            # perf script and trace-cmd report puts the event name on fifth
            # field.
            # perf:
            #   kdamond.0  4452 [000] 82877.315633: damon:damon_aggregated: ...
            # trace-cmd:
            #   kdamond.0-264454 [007] ..... 92627.258073: damon_aggregated: ...
            if len(fields) < 5:
                continue
            event = fields[4][:-1]
        elif trace_text_format == 'damo-report-trace-perf':
            # 0.000 kdamond.0/83017 damon:damon_aggregated(nr_regions: 11, ...
            if len(fields) < 3:
                continue
            event = fields[2].split('(')[0]
        elif trace_text_format == 'damo-report-trace-trace-cmd':
            # <...>-83432 [006] ..... 71629.102757: damon_aggregated: ...
            if len(fields) < 5:
                continue
            event = 'damon:%s' % fields[4][:-1]

        if not event in events:
            continue
        pr_trace_line(line, args.raw, trace_text_format, args.max_cols)

def main(args):
    if args.input is not None:
        return report_recorded_trace(args)

    global tracer_pipe

    _damon.ensure_root_permission()

    if args.event is None:
        print('--event is required')
        exit(1)

    events = get_events_to_show(args.event, args.no_event)

    if args.tracer is None:
        if _damo_subproc.avail_cmd('perf'):
            tracer = 'perf'
        elif _damo_subproc.avail_cmd('trace-cmd'):
            tracer = 'trace-cmd'
    else:
        tracer = args.tracer

    if tracer == 'perf':
        cmd = [tracer, 'trace', '--libtraceevent_print']
    else:
        cmd = [tracer, 'stream']

    for trace_event in events:
        cmd.append('-e')
        cmd.append(trace_event)

    signal.signal(signal.SIGINT, signalhandler)
    signal.signal(signal.SIGTERM, signalhandler)

    global output_file
    if args.output is not None:
        output_file = open(args.output, 'w')
        output_file.write('damo_report_trace_output_format_version: 0\n')
        output_file.write('damo_report_trace_output_time: %s\n' %
                          datetime.datetime.now().timestamp())
        output_file.write('damo_report_trace_output_tracer: %s\n' % tracer)

    tracer_pipe = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    while True:
        output = tracer_pipe.stdout.readline()
        if not output and tracer_pipe.poll() is not None:
            break
        output = output.decode()
        pr_trace_line(output, args.raw, 'damo-report-trace-%s' % tracer,
                      args.max_cols)
        if output_file is not None:
            output_file.write(output)

def set_argparser(parser):
    parser.add_argument(
            '--event', choices=list(damon_trace_events) + ['all'], nargs='+',
            default='all',
            help='events to trace')
    parser.add_argument(
            '--no_event', choices=list(damon_trace_events), nargs='+',
            help='events to skip tracing')
    parser.add_argument('--tracer', choices=['perf', 'trace-cmd'],
                        help='tracer command to use')
    parser.add_argument('--input', '-i', metavar='<file>',
                        help='trace record file')
    parser.add_argument('--output', '-o', metavar='<file>',
                        help='save output to a file')
    parser.add_argument('--raw', action='store_true',
                        help='show raw trace output')
    parser.add_argument('--max_cols', type=int, default=100,
                        help='max columns per line')



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import unittest

import _test_damo_common

_test_damo_common.add_damo_dir_to_syspath()

import damo_report_trace

class TestDamoReportTrace(unittest.TestCase):
    def test_get_trace_fields(self):
        self.assertEqual(damo_report_trace.get_trace_fields(
            '2452.789 kdamond.0/2149 damon:damon_region_aggregated(target_id=0 nr_regions=11 8354394112-8372879360: 0 600 probe_hits=13 00)'.split(),
            'damo-report-trace-perf', 'damon:damon_region_aggregated'),
            'target_id=0 nr_regions=11 8354394112-8372879360: 0 600 probe_hits=13 00'.split())

    def test_fmt_damon_region_aggregated(self):
        self.assertEqual(damo_report_trace.fmt_damon_region_aggregated(
            '2452.789 kdamond.0/2149 damon:damon_region_aggregated(target_id=0 nr_regions=11 8354394112-8372879360: 0 600 probe_hits=13 00)'.split(), 'damo-report-trace-perf'),
                         '2452.789 kdamond.0/2149 damon_region_aggregated 0 0/11 7.780636 GiB (17.628906 MiB) 0 600 19 0')

    def test_fmt_damon_region_aggregated_trace(self):
        damo_report_trace.region_aggregated_idx = 0
        self.assertEqual(damo_report_trace.fmt_damon_region_aggregated_trace(
            'target_id=0 nr_regions=11 8354394112-8372879360: 0 600 probe_hits=13 00'.split()),
            '0 0/11 7.780636 GiB (17.628906 MiB) 0 600 19 0')

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0

import sys

contact_message = '''
    If you depend on those, please report your usecase to GitHub Issues [1],
    sj@kernel.org, damon@lists.linux.dev and/or linux-mm@kvack.org.

    [1] https://github.com/damonitor/damo/issues

'''

def will_be_deprecated(feature, deadline, additional_notice='',
                       alternative=None):
    lines = [
            '',
            'WARNING: %s will be deprecated by %s.' % (feature, deadline)
            ]
    if alternative:
        lines.append('    Use "%s" instead.' % alternative)
    if additional_notice:
        lines.append(additional_notice)
    lines.append(contact_message)
    sys.stderr.write('\n'.join(lines))

def deprecated(feature, deadline, alternative=None, do_exit=False, exit_code=1,
        additional_notice=''):
    lines = [
            'WARNING: %s is deprecated.' % feature,
            '    The support will be removed by %s.' % deadline,
            ]
    if alternative:
        lines.append('    Use "%s" instead.' % alternative)
    if additional_notice:
        lines.append(additional_notice)
    lines.append(contact_message)
    sys.stderr.write('\n'.join(lines))
    if do_exit:
        exit(exit_code)



# SPDX-License-Identifier: GPL-2.0

"""
Start DAMON with given parameters.
"""

import os
import signal
import time

import _damo_records
import _damo_sysinfo
import _damon
import _damon_args
import _damon_modules

def handle_modules():
    sysinfo = _damo_sysinfo.get_sysinfo_or_panic()
    for module in os.listdir(os.path.join(sysinfo.sysfs_path, 'module')):
        if not module.startswith('damon_'):
            continue
        if not _damon_modules.module_running(module):
            continue
        print('Cannot turn on damon since %s is running.  '
              'You should disable it first.' % module)
        answer = input('May I disable it for you? [Y/n] ')
        if answer.lower() == 'n':
            print('Ok, see you later')
            exit(1)
        print('Ok, disabling it')
        _damon_modules.module_disable(module)
        print('Disabled it.  Continue starting DAMON')

def sighandler(signum, frame):
    print('\nsingal %s received' % signum)
    exit(0)

def main(args):
    _damon.ensure_root_and_initialized(args)
    handle_modules()

    err, kdamonds = _damon_args.turn_damon_on(args)
    if err:
        print('could not turn on damon (%s)' % err)
        exit(1)

    if args.include_child_tasks is True:
        signal.signal(signal.SIGINT, sighandler)
        signal.signal(signal.SIGTERM, sighandler)
        print('Continue monitoring child tasks and updating DAMON targets')
        print('Press Ctrl+C to stop')
        while True:
            _damon.add_commit_vaddr_child_targets(kdamonds)
            time.sleep(3)

def set_argparser(parser):
    _damon_args.set_argparser(parser, add_record_options=False, min_help=True)
    parser.add_argument('--include_child_tasks', action='store_true',
                        help='add child tasks as monitoring target')
    parser.description = 'Start DAMON with specified parameters'
    return parser



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import unittest

import _test_damo_common

_test_damo_common.add_damo_dir_to_syspath()

import _damo_fs
import _damon_sysfs

class TestDamonSysfs(unittest.TestCase):
    def test_json_kdamonds_staging(self):
        sysfs_dict = {
                "nr_kdamonds": "1\n",
                "0": {
                    "state": "off\n",
                    "pid": "-1\n",
                    "avail_state_inputs": '\n'.join(['on', 'off', 'commit',
                        'update_schemes_stats', 'update_schemes_tried_regions',
                        'clear_schemes_tried_regions\n']),
                    "contexts": {
                        "nr_contexts": "1\n",
                        "0": {
                            "avail_operations": "vaddr\nfvaddr\npaddr\n",
                            "operations": "paddr\n",
                            "monitoring_attrs": {
                                "intervals": {
                                    "sample_us": "5000\n",
                                    "update_us": "1000000\n",
                                    "aggr_us": "200000\n",
                                    },
                                "nr_regions": {
                                    "max": "1000\n", "min": "10\n"}
                                },
                            "targets": {"nr_targets": "0\n"},
                            "schemes": {
                                "nr_schemes": "1\n",
                                "0": {
                                    "access_pattern": {
                                        "age": {"max": "0\n", "min": "0\n"},
                                        "nr_accesses":
                                        {"max": "0\n", "min": "0\n"},
                                        "sz": {"max": "0\n", "min": "0\n"}},
                                    "action": "stat\n",
                                    "quotas": {
                                        "bytes": "0\n", "ms": "0\n",
                                        "reset_interval_ms": "0\n",
                                        "weights": {
                                            "age_permil": "0\n",
                                            "nr_accesses_permil": "0\n",
                                            "sz_permil": "0\n"
                                            }
                                        },
                                    "watermarks": {
                                        "metric": "none\n", "high": "0\n",
                                        "mid": "0\n", "low": "0\n",
                                        "interval_us": "0\n",
                                        },
                                    "stats": {
                                        "nr_applied": "0\n", "nr_tried": "0\n",
                                        "sz_applied": "0\n", "sz_tried": "0\n",
                                        "qt_exceeds": "0\n",
                                        },
                                    "tried_regions": {},
                                   },
                                },
                            },
                        },
                    }
                }
        expected_dryrun_log = [
                "read 'nr_kdamonds': '0'",
                "write '1' to 'nr_kdamonds'",
                "read '0/contexts/nr_contexts': '0'",
                "write '1' to '0/contexts/nr_contexts'",
                "write 'paddr' to '0/contexts/0/operations'",
                "write '5000' to '0/contexts/0/monitoring_attrs/intervals/sample_us'",
                "write '200000' to '0/contexts/0/monitoring_attrs/intervals/aggr_us'",
                "write '1000000' to '0/contexts/0/monitoring_attrs/intervals/update_us'",
                "write '10' to '0/contexts/0/monitoring_attrs/nr_regions/min'",
                "write '1000' to '0/contexts/0/monitoring_attrs/nr_regions/max'",
                "read '0/contexts/0/targets/nr_targets': '0'",
                "read '0/contexts/0/schemes/nr_schemes': '0'",
                "write '1' to '0/contexts/0/schemes/nr_schemes'",
                "write '0' to '0/contexts/0/schemes/0/access_pattern/sz/min'",
                "write '0' to '0/contexts/0/schemes/0/access_pattern/sz/max'",
                "write '0' to '0/contexts/0/schemes/0/access_pattern/nr_accesses/min'",
                "write '0' to '0/contexts/0/schemes/0/access_pattern/nr_accesses/max'",
                "write '0' to '0/contexts/0/schemes/0/access_pattern/age/min'",
                "write '0' to '0/contexts/0/schemes/0/access_pattern/age/max'",
                "write 'stat' to '0/contexts/0/schemes/0/action'",
                "write '0' to '0/contexts/0/schemes/0/quotas/ms'",
                "write '0' to '0/contexts/0/schemes/0/quotas/bytes'",
                "write '0' to '0/contexts/0/schemes/0/quotas/reset_interval_ms'",
                "write '0' to '0/contexts/0/schemes/0/quotas/weights/sz_permil'",
                "write '0' to '0/contexts/0/schemes/0/quotas/weights/nr_accesses_permil'",
                "write '0' to '0/contexts/0/schemes/0/quotas/weights/age_permil'",
                "write 'none' to '0/contexts/0/schemes/0/watermarks/metric'",
                "write '0' to '0/contexts/0/schemes/0/watermarks/interval_us'",
                "write '0' to '0/contexts/0/schemes/0/watermarks/high'",
                "write '0' to '0/contexts/0/schemes/0/watermarks/mid'",
                "write '0' to '0/contexts/0/schemes/0/watermarks/low'",
                ]

        _damo_fs.debug_dryrun(
                {
                    'nr_kdamonds': '0',
                    '0/contexts/nr_contexts': '0',
                    '0/contexts/0/targets/nr_targets': '0',
                    '0/contexts/0/schemes/nr_schemes': '0',
                    })
        kdamonds = _damon_sysfs.files_content_to_kdamonds(sysfs_dict)
        _damon_sysfs.write_kdamonds_dir('', kdamonds)
        logs = _damo_fs.debug_get_dryrun_logs()
        self.assertEqual(expected_dryrun_log, logs)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import json
import unittest

import _test_damo_common

_test_damo_common.add_damo_dir_to_syspath()

import _damo_fs
import _damo_sysinfo
import _damon
import _damon_dbgfs
import _damon_features

def set_damon_dbgfs_features():
    avail_features = []
    for name, support in {
            'debugfs/init_regions': True, 'debugfs/schemes': True,
            'debugfs/schemes_stat_qt_exceed': True,
            'debugfs/init_regions_target_idx': True,
            'debugfs/schemes_prioritization': True,
            'sysfs/schemes_tried_regions': False, 'debugfs/record': False,
            'debugfs/schemes_time_quota': True,
            'sysfs/schemes_time_quota': True,
            'sysfs/fvaddr': False, 'debugfs/paddr': True,
            'debugfs/schemes_wmarks': True, 'debugfs/schemes_size_quota': True,
            'debugfs/schemes_stat_succ': True, 'debugfs/vaddr': True}.items():
        if not support:
            continue
        for feature in _damon_features.features_list:
            if feature.name == name:
                avail_features.append(feature)
    _damo_sysinfo.system_info = _damo_sysinfo.SystemInfo(
            damo_version=None, kernel_version=None,
            avail_damon_features=avail_features)

class TestDamonDbgfs(unittest.TestCase):
    def test_nr_kdamonds(self):
        self.assertEqual(_damon_dbgfs.nr_kdamonds(), 1)

    def test_debugfs_output_to_damos(self):
        set_damon_dbgfs_features()

        _test_damo_common.test_input_expects(self,
                lambda x: _damon_dbgfs.debugfs_output_to_damos(
                    x, _damon.DamonIntervals(5000, 100000, 1000000)),
                {"4096 18446744073709551615 0 0 10 42949 5 0 584792941 1000 0 0 0 0 0 0 0 0 0 0 0 0 0\n":
                    _damon.Damos(
                        access_pattern=_damon.DamosAccessPattern(
                            [4096, 18446744073709551615],
                            [0.0, 0.0], _damon.unit_percent,
                            [1000000.0, 4294900000.0], _damon.unit_usec),
                        action='stat', target_nid=None,
                        apply_interval_us=None,
                        quotas=_damon.DamosQuotas(time_ms=0,
                            sz_bytes=584792941, reset_interval_ms=1000,
                            weights=[0,0,0]),
                        watermarks=_damon.DamosWatermarks('none',0,0,0,0),
                        filters=[], stats=None)})

    def test_files_content_to_kdamonds_io(self):
        set_damon_dbgfs_features()

        _damo_fs.debug_dryrun({})
        kdamonds = _damon_dbgfs.files_content_to_kdamonds(json.loads(
            json.dumps(
                {
                    "attrs": "5000 100000 1000000 10 1000\n",
                    "init_regions": "0 1 100\n0 100 200\n",
                    "kdamond_pid": "none\n",
                    "mk_contexts":
                    "read failed (reading /sys/kernel/debug/damon/mk_contexts failed ([Errno 22] Invalid argument))",
                    "monitor_on": "off\n",
                    "rm_contexts":
                    "read failed (reading /sys/kernel/debug/damon/rm_contexts failed ([Errno 22] Invalid argument))",
                    "schemes":
                    "4096 18446744073709551615 0 0 10 42949 5 0 584792941 1000 0 0 0 0 0 0 0 0 0 0 0 0 0\n",
                    "target_ids": "4242\n"
                    })))
        _damon_dbgfs.write_kdamonds('', kdamonds)

        kdamonds = _damon_dbgfs.files_content_to_kdamonds(json.loads(
            json.dumps(
                {
                    "attrs": "5000 100000 1000000 10 1000\n",
                    "init_regions": "",
                    "kdamond_pid": "none\n",
                    "mk_contexts":
                    "read failed (reading /sys/kernel/debug/damon/mk_contexts failed ([Errno 22] Invalid argument))",
                    "monitor_on": "off\n",
                    "rm_contexts":
                    "read failed (reading /sys/kernel/debug/damon/rm_contexts failed ([Errno 22] Invalid argument))",
                    "schemes":
                    "4096 18446744073709551615 0 0 10 42949 5 0 5368709120 1000 0 3 7 1 1000000 999 998 995 0 0 0 0 0\n",
                    "target_ids": "42\n"
                    })))
        _damon_dbgfs.write_kdamonds('', kdamonds)

        logs = _damo_fs.debug_get_dryrun_logs()
        expected_logs = [
                # First write_kdamonds()
                "write '5000 100000 1000000 10 1000' to 'attrs'",
                "write '4242' to 'target_ids'",
                "write '0 1 100 0 100 200' to 'init_regions'",
                "write '4096\t18446744073709551615\t0\t0\t10\t42949\t5\t0\t584792941\t1000\t0\t0\t0\t0\t0\t0\t0\t0' to 'schemes'",

                # Second write_kdamonds()
                "write '5000 100000 1000000 10 1000' to 'attrs'",
                "write 'paddr' to 'target_ids'",
                "write '' to 'init_regions'",
                "write '4096\t18446744073709551615\t0\t0\t10\t42949\t5\t0\t5368709120\t1000\t0\t3\t7\t1\t1000000\t999\t998\t995' to 'schemes'",
                ]
        self.assertEqual(expected_logs, logs)

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0

'''
Contains code for managing system information including kernel version and
DAMON features enabled on the kernel.
'''

import collections
import json
import os
import subprocess

import _damo_fs
import _damo_subproc
import _damon_dbgfs
import _damon_features
import _damon_modules
import _damon_sysfs
import damo_version

class SystemInfo:
    damo_version = None
    kernel_version = None

    sysfs_path = None
    tracefs_path = None
    debugfs_path = None

    trace_cmd_version = None
    perf_path = None
    perf_version = None

    # DAMON features that available on current kernel.
    avail_damon_features = None

    def __init__(self, damo_version, kernel_version,
                 perf_path=None, perf_version=None, trace_cmd_version=None,
                 sysfs_path=None, tracefs_path=None, debugfs_path=None,
                 avail_damon_features=None):
        self.damo_version = damo_version
        self.kernel_version = kernel_version

        self.sysfs_path = sysfs_path
        self.tracefs_path = tracefs_path
        self.debugfs_path = debugfs_path

        self.trace_cmd_version=trace_cmd_version
        self.perf_path = perf_path
        self.perf_version = perf_version

        self.avail_damon_features = avail_damon_features

    def to_kvpairs(self, raw=False):
        return collections.OrderedDict([
            ('damo_version', self.damo_version),
            ('kernel_version', self.kernel_version),
            ('sysfs_path', self.sysfs_path),
            ('tracefs_path', self.tracefs_path),
            ('debugfs_path', self.debugfs_path),
            ('trace_cmd_version', self.trace_cmd_version),
            ('perf_path', self.perf_path),
            ('perf_version', self.perf_version),
            ('avail_damon_features',
             [f.to_kvpairs(raw) for f in self.avail_damon_features]),
            ])

    @classmethod
    def from_kvpairs(cls, kvpairs):
        trace_cmd_version = None
        if 'trace_cmd_version' in kvpairs:
            trace_cmd_version = kvpairs['trace_cmd_version']
        perf_path = None
        if 'perf_path' in kvpairs:
            perf_path = kvpairs['perf_path']
        perf_version = None
        if 'perf_version' in kvpairs:
            perf_version = kvpairs['perf_version']
        avail_damon_features = []
        if 'avail_damon_features' in kvpairs:
            avail_damon_features = kvpairs['avail_damon_features']
        sysfs_path = None
        if 'sysfs_path' in kvpairs:
            sysfs_path = kvpairs['sysfs_path']
        tracefs_path = None
        if 'tracefs_path' in kvpairs:
            tracefs_path = kvpairs['tracefs_path']
        debugfs_path = None
        if 'debugfs_path' in kvpairs:
            debugfs_path = kvpairs['debugfs_path']
        return cls(
                damo_version=kvpairs['damo_version'],
                kernel_version=kvpairs['kernel_version'],
                sysfs_path=sysfs_path,
                tracefs_path=tracefs_path,
                debugfs_path=debugfs_path,
                trace_cmd_version=trace_cmd_version,
                perf_path=perf_path, perf_version=perf_version,
                avail_damon_features=[
                    _damon_features.DamonFeature.from_kvpairs(kvp) for kvp in
                    avail_damon_features],
                )

    def __eq__(self, other):
        return self.damo_version == other.damo_version and \
                self.kernel_version == other.kernel_version and \
                self.sysfs_path == other.sysfs_path and \
                self.tracefs_path == other.tracefs_path and \
                self.debugfs_path == other.debugfs_path and \
                self.trace_cmd_version == other.trace_cmd_version and \
                self.perf_path == other.perf_path and \
                self.perf_version == other.perf_version and \
                self.avail_damon_features == other.avail_damon_features

    def feature_available(self, feature_name):
        return feature_name in [f.name for f in self.avail_damon_features]

    def infer_damon_version(self):
        '''Return version string and error'''
        avail_features = {f.name for f in self.avail_damon_features}
        if len(avail_features) == 0:
            return '<5.15', None
        append_plus = False
        for feature in reversed(_damon_features.features_list):
            if feature.name in avail_features:
                if feature.upstreamed_version in ['none', 'unknown']:
                    append_plus = True
                else:
                    version = feature.upstreamed_version
                    if append_plus:
                        version = '%s+' % version
                    return version, None
        return None, 'only non-upstreamed features'

system_info = None
sysinfo_file_path = os.path.join(os.environ['HOME'], '.damo.sysinfo')

def read_sysinfo_file():
    '''
    Read save_sysinfo_file()-saved SystemInfo object.
    Returns read SystemInfo object and an error string if failed.
    '''
    if not os.path.isfile(sysinfo_file_path):
        return None, 'sysinfo file (%s) not found' % sysinfo_file_path
    try:
        with open(sysinfo_file_path, 'r') as f:
            kvpairs = json.load(f)
    except Exception as e:
        return None, 'json load of %s failed' % sysinfo_file_path
    return SystemInfo.from_kvpairs(kvpairs), None

def valid_cached_sysinfo(sysinfo, damo_version_, kernel_version):
    if sysinfo.damo_version != damo_version_:
        return False
    if sysinfo.kernel_version != kernel_version:
        return False
    return True

def avail_features_on(damon_fs):
    if not damon_fs.supported():
        return [], None
    feature_supports_map, err = damon_fs.mk_feature_supports_map()
    if err is not None:
        return None, 'feature map making fail (%s)' % err
    avail_features = [f for f in _damon_features.features_list
                      if feature_supports_map[f.name]]
    return avail_features, None

def get_damon_tracepoints():
    '''
    returns list of DAMON tracepoint names and an error
    '''
    tracefs_path = _damo_fs.dev_mount_point('tracefs')
    if tracefs_path is None:
        return None, 'tracefs is not mounted'
    points = []
    try:
        with open(os.path.join(tracefs_path, 'available_events'), 'r') as f:
            for line in f:
                if line.startswith('damon:'):
                    points.append(line.strip())
    except Exception as e:
        return None, 'available tracepoints reading fail'
    return points, None

tracepoint_to_feature_name_map = {
        'damon:damon_aggregated': 'trace/damon_aggregated',
        'damon:damos_before_apply': 'trace/damos_before_apply',
        'damon:damon_monitor_intervals_tune':
        'trace/damon_monitor_intervals_tune',
        'damon:damos_esz': 'trace/damos_esz',
        'damon:damos_stat_after_apply_interval':
        'trace/damos_stat_after_apply_interval',
        'damon:damon_region_aggregated': 'trace/damon_region_aggregated',
        }

def get_avail_damon_trace_features():
    features = []
    tracepoints, err = get_damon_tracepoints()
    if err is not None:
        return None, err
    for tracepoint, feature_name in tracepoint_to_feature_name_map.items():
        if tracepoint in tracepoints:
            features.append(_damon_features.feature_of_name(feature_name))
    return features, err

def get_avail_damon_interface_features():
    features = []

    if _damon_dbgfs.supported():
        features.append(_damon_features.feature_of_name(
            'interface/damon_debugfs'))

    sysfs_path = _damo_fs.dev_mount_point('sysfs')
    if sysfs_path is None:
        return features

    if _damon_sysfs.supported():
        features.append(_damon_features.feature_of_name(
            'interface/damon_sysfs'))

    features += _damon_modules.get_avail_interface_features()

    return features

def get_trace_cmd_version():
    if not _damo_subproc.avail_cmd('trace-cmd'):
        return None
    try:
        output = subprocess.check_output(['trace-cmd']).decode().strip()
    except Exception as e:
        output = e.output.decode().strip()
    for line in output.split('\n'):
        # version line is, e.g., trace-cmd version 3.3.1 (not-a-git-repo)
        if not line.startswith('trace-cmd version '):
            continue
        fields = line.split()
        return ' '.join(fields[2:])
    return None

def get_perf_path_version():
    try:
        perf_path = subprocess.check_output(['which', 'perf']).decode().strip()
    except:
        perf_path = None
    if perf_path is not None:
        perf_version = subprocess.check_output(
                ['perf', '--version']).decode().strip()
    else:
        perf_version = None
    return perf_path, perf_version

def get_sysinfo_from_scratch():
    damo_version_ = damo_version.get_real_version()
    kernel_version = subprocess.check_output(['uname', '-r']).decode().strip()
    sysfs_path = _damo_fs.dev_mount_point('sysfs')
    tracefs_path = _damo_fs.dev_mount_point('tracefs')
    debugfs_path = _damo_fs.dev_mount_point('debugfs')

    trace_cmd_version = get_trace_cmd_version()
    perf_path, perf_version = get_perf_path_version()

    avail_damon_features = []
    avail_damon_sysfs_features, err = avail_features_on(_damon_sysfs)
    if err is not None:
        return None, 'sysfs feature check fail (%s)' % err
    avail_damon_features += avail_damon_sysfs_features

    avail_damon_debugfs_features, err = avail_features_on(_damon_dbgfs)
    if err is not None:
        return None, 'debugfs feature check fail (%s)' % err
    avail_damon_features += avail_damon_debugfs_features

    avail_damon_trace_features, err = get_avail_damon_trace_features()
    if err is not None:
        return None, 'trace feature check fail (%s)' % err
    avail_damon_features += avail_damon_trace_features

    avail_damon_features += get_avail_damon_interface_features()
    avail_damon_features += _damon_modules.get_avail_features()

    sysinfo = SystemInfo(
            damo_version=damo_version_,
            kernel_version=kernel_version,
            sysfs_path=sysfs_path,
            tracefs_path=tracefs_path,
            debugfs_path=debugfs_path,
            trace_cmd_version=trace_cmd_version,
            perf_path=perf_path, perf_version=perf_version,
            avail_damon_features=avail_damon_features)
    return sysinfo, None

def version_mismatch(sysinfo):
    if sysinfo.damo_version != damo_version.get_real_version():
        return True
    kernel_version = subprocess.check_output(['uname', '-r']).decode().strip()
    if sysinfo.kernel_version != kernel_version:
        return True
    return False

def update_cached_info(cached_info):
    if cached_info is None:
        return get_sysinfo_from_scratch()
    if version_mismatch(cached_info):
        return get_sysinfo_from_scratch()

    trace_cmd_version = get_trace_cmd_version()
    perf_path, perf_version = get_perf_path_version()

    sysfs_path = _damo_fs.dev_mount_point('sysfs')
    if cached_info.sysfs_path != sysfs_path:
        cached_info.sysfs_path = sysfs_path
        avail_damon_sysfs_features, err = avail_features_on(_damon_sysfs)
        if err is not None:
            return None, 'damon sysfs features update fail (%s)' % err

        avail_damon_features = []
        for f in cached_info.avail_damon_features:
            interfaces_to_recheck = ['sysfs', 'interface', 'reclaim',
                                     'lru_sort', 'stat']
            skip = False
            for interface in interfaces_to_recheck:
                if f.name.startswith('%s/' % interface):
                    skip = True
                    break
            if skip:
                continue
            avail_damon_features.append(f)
        avail_damon_features += avail_damon_sysfs_feastures
        avail_damon_features += get_avail_damon_interface_features()
        avail_damon_features += _damon_modules.get_avail_features()
        cached_info.avail_damon_features = avail_damon_features

    tracefs_path = _damo_fs.dev_mount_point('tracefs')
    if cached_info.tracefs_path != tracefs_path:
        cached_info.tracefs_path = tracefs_path
        avail_damon_trace_features, err = get_avail_damon_trace_features()
        if err is not None:
            return None, 'damon trace features update fail (%s)' % err

        avail_damon_features = []
        for f in cached_info.avail_damon_features:
            if f.name.startswith('trace/'):
                continue
            avail_damon_features.append(f)
        avail_damon_features += avail_damon_trace_features
        cached_info.avail_damon_features = avail_damon_features

    debugfs_path = _damo_fs.dev_mount_point('debugfs')
    if cached_info.debugfs_path != debugfs_path:
        cached_info.debugfs_path = debugfs_path
        avail_damon_debugfs_features, err = avail_features_on(_damon_debugfs)
        if err is not None:
            return None, 'damon debugfs features update fail (%s)' % err

        avail_damon_features = []
        for f in cached_info.avail_damon_features:
            if f.name.startswith('debugfs/'):
                continue
            avail_damon_features.append(f)
        avail_damon_features += avail_damon_debugfs_features
        cached_info.avail_damon_features = avail_damon_features

    return cached_info, None

def load_sysinfo():
    '''
    Set system_info global variable.

    Returns an error if failed.
    '''
    cached_info, cache_read_err = read_sysinfo_file()
    info, err = update_cached_info(cached_info)
    if err is not None:
        errs = []
        if cache_read_err is not None:
            errs.append('cache read fail (%s)' % cache_read_err)
        errs.append('info update fail (%s)' % err)
        return 'sysinfo loading fail (%s)' % ', '.join(errs)

    global system_info
    system_info = info
    save_sysinfo_file()
    return None

def save_sysinfo_file():
    '''
    Save system_info as a file that we can read later.

    Returns error in case of failure.
    '''
    if system_info is None:
        return 'system_info is not initialized'
    try:
        with open(sysinfo_file_path, 'w') as f:
            json.dump(system_info.to_kvpairs(), f, indent=4)
    except Exception as e:
        return 'json dump fail (%s)' % e
    return None

def rm_sysinfo_file():
    try:
        os.remove(sysinfo_file_path)
    except Exception as e:
        return '%s' % e
    return None

def get_sysinfo():
    if system_info is None:
        err = load_sysinfo()
        if err is not None:
            return None, err
    return system_info, None

def get_sysinfo_or_panic():
    sysinfo, err = get_sysinfo()
    if err is not None:
        raise Exception('[PANIC] get_sysinfo() fail (%s)' % err)
        exit(1)
    return sysinfo

def damon_tracepoint_available(tracepoint):
    sysinfo, err = get_sysinfo()
    if err is not None:
        return False
    feature_name = tracepoint_to_feature_name_map[tracepoint]
    return feature_name in [f.name for f in sysinfo.avail_damon_features]

def damon_feature_available(feature_name):
    sysinfo, err = get_sysinfo()
    if err is not None:
        return False
    for f in sysinfo.avail_damon_features:
        if f.name == feature_name:
            return True
    return False

def available_damon_tracepoints():
    sysinfo, err = get_sysinfo()
    if err is not None:
        return []
    tracepoints = []
    for f in sysinfo.avail_damon_features:
        if f.name.startswith('trace/'):
            tracepoints.append('damon:%s' % f.name[6:])
    return tracepoints



# SPDX-License-Identifier: GPL-2.0

"""
Contains core functions for DAMON sysfs control.
"""

import os
import time

import _damo_fmt_str
import _damo_fs
import _damon
import _damon_features

sysfs_root = None

def get_sysfs_root():
    global sysfs_root

    if sysfs_root is None:
        sysfs_root = _damo_fs.dev_mount_point('sysfs')
    return sysfs_root

def get_kdamonds_dir():
    '''Returns None if sysfs is not mounted'''
    if get_sysfs_root() is None:
        return None
    return os.path.join(get_sysfs_root(), 'kernel/mm/damon/admin/kdamonds')

def supported():
    kdamonds_dir = get_kdamonds_dir()
    if kdamonds_dir is None:
        return False
    return os.path.isdir(kdamonds_dir)

def get_state_file_of(kdamond_idx):
    return os.path.join(get_kdamonds_dir(), '%s' % kdamond_idx, 'state')

def __write_state_file(kdamond_idxs, command):
    'Return error'
    err = None
    for kdamond_idx in kdamond_idxs:
        err = _damo_fs.write_file(get_state_file_of(kdamond_idx), command)
        if err != None:
            break
    return err

def turn_damon_on(kdamonds_idxs):
    # In case of vaddr, too early monitoring shows unstable mapping changes.
    # Give the process a time to have stable memory mapping.
    time.sleep(0.5)
    return __write_state_file(kdamonds_idxs, 'on')

def turn_damon_off(kdamonds_idxs):
    return __write_state_file(kdamonds_idxs, 'off')

def is_kdamond_running(kdamond_idx):
    content, err = _damo_fs.read_file(get_state_file_of(kdamond_idx))
    if err != None:
        print(err)
        return False
    return content.strip() == 'on'

def get_refresh_ms_file_of(kdamond_idx):
    return os.path.join(get_kdamonds_dir(), '%s' % kdamond_idx, 'refresh_ms')

def refresh_ms_enabled(kdamond_idx):
    filepath = get_refresh_ms_file_of(kdamond_idx)
    if not os.path.isfile(filepath):
        return False, None
    content, err = _damo_fs.read_file(get_refresh_ms_file_of(kdamond_idx))
    if err is not None:
        return False, err
    try:
        enabled = int(content) != 0
    except Exception as e:
        return False, 'int(refresh_ms) failed (%s)!  kernel bug?' % e
    return enabled, None

def refresh_ms_disabled_kdidxs(kdamond_idxs):
    idxs = []
    for kdidx in kdamond_idxs:
        unnecessary, err = refresh_ms_enabled(kdidx)
        if err is not None:
            return None, 'refresh_ms check fail (%s)' % err
        if not unnecessary:
            idxs.append(kdidx)
    return idxs, None

'Return error'
def update_tuned_intervals(kdamond_idxs):
    kdamond_idxs, err = refresh_ms_disabled_kdidxs(kdamond_idxs)
    if err is not None:
        return err
    return __write_state_file(kdamond_idxs, 'update_tuned_intervals')

'Return error'
def update_schemes_stats(kdamond_idxs):
    kdamond_idxs, err = refresh_ms_disabled_kdidxs(kdamond_idxs)
    if err is not None:
        return err
    return __write_state_file(kdamond_idxs, 'update_schemes_stats')

'Return error'
def update_schemes_tried_bytes(kdamond_idxs):
    err = __write_state_file(kdamond_idxs, 'update_schemes_tried_bytes')
    if err != None:
        err = '%s (maybe schemes_tried_regions_sz not supported?)' % err
    return err

'Return error'
def update_schemes_tried_regions(kdamond_idxs):
    return __write_state_file(kdamond_idxs, 'update_schemes_tried_regions')

'Return error'
def update_schemes_quota_effective_bytes(kdamond_idxs):
    kdamond_idxs, err = refresh_ms_disabled_kdidxs(kdamond_idxs)
    if err is not None:
        return err
    err = __write_state_file(kdamond_idxs, 'update_schemes_effective_quotas')
    if err != None:
        err = '%s (maybe schemes_effective_quotas not supported?)' % err
    return err

# for stage_kdamonds

def write_filter_dir(dir_path, filter_):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'type'), filter_.filter_type)
    if err is not None:
        # todo: make error message more detailed.
        # anon/memcg are merged in 6.3-rc1
        # addr/target are merged in 6.6-rc1
        return err

    if filter_.allow is True:
        if os.path.isfile(os.path.join(dir_path, 'allow')):
            err = _damo_fs.write_file(os.path.join(dir_path, 'allow'),
                                      'Y' if filter_.allow else 'N')
        # pass file has renamed to allow during development
        elif os.path.isfile(os.path.join(dir_path, 'pass')):
            err = _damo_fs.write_file(os.path.join(dir_path, 'pass'),
                                      'Y' if filter_.allow else 'N')
        else:
            return 'kernel is not supporting allow_filter'
        if err is not None:
            return err

    if filter_.memcg_path is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'memcg_path'), filter_.memcg_path)
        if err is not None:
            return err

    if filter_.hugepage_size is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'min'),
                '%d' % filter_.hugepage_size[0])
        if err is not None:
            return err

        err = _damo_fs.write_file(
                os.path.join(dir_path, 'max'),
                '%d' % filter_.hugepage_size[1])
        if err is not None:
            return err

    if filter_.address_range is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'addr_start'),
                '%d' % filter_.address_range.start)
        if err is not None:
            return err

        err = _damo_fs.write_file(
                os.path.join(dir_path, 'addr_end'),
                '%d' % filter_.address_range.end)
        if err is not None:
            return err

    if filter_.damon_target_idx is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'damon_target_idx'),
                '%d' % filter_.damon_target_idx)
        if err is not None:
            return err

    return _damo_fs.write_file(os.path.join(dir_path, 'matching'),
                               'Y' if filter_.matching else 'N')

def ensure_nr_file_for(file_path, list_):
    content, err = _damo_fs.read_file(file_path)
    if err is not None:
        return err
    current_nr = int(content)
    desired_nr = len(list_)
    if current_nr == desired_nr:
        return None
    return _damo_fs.write_file(file_path, '%d' % desired_nr)

def write_filters_dir(dir_path, filters):
    # filters merged in v6.3-rc1
    if not os.path.isdir(dir_path):
        if len(filters) == 0:
            return None
        return 'the kernel is not supporting filters'

    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_filters'), filters)
    if err is not None:
        return err

    for idx, filter_ in enumerate(filters):
        err = write_filter_dir(os.path.join(dir_path, '%d' % idx), filter_)
        if err is not None:
            return err
    return None

def is_core_filter(damos_filter):
    return damos_filter.filter_type in ['addr', 'target']

def write_core_ops_filters_dirs(scheme_dir_path, filters):
    core_filters = [f for f in filters if is_core_filter(f)]
    err = write_filters_dir(os.path.join(scheme_dir_path, 'core_filters'),
                            core_filters)
    if err is not None:
        return err

    ops_filters = [f for f in filters if not is_core_filter(f)]
    err = write_filters_dir(os.path.join(scheme_dir_path, 'ops_filters'),
                            ops_filters)
    if err is not None:
        return err

def write_filters_dirs(scheme_dir_path, filters):
    if os.path.isdir(os.path.join(scheme_dir_path, 'core_filters')):
        return write_core_ops_filters_dirs(scheme_dir_path, filters)
    return write_filters_dir(os.path.join(scheme_dir_path, 'filters'), filters)

def write_watermarks_dir(dir_path, wmarks):
    if wmarks is None:
        # TODO: ensure wmarks is not None
        return None
    err = _damo_fs.write_file(os.path.join(dir_path, 'metric'), wmarks.metric)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'interval_us'), '%d' % wmarks.interval_us)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'high'), '%d' % wmarks.high_permil)
    if err is not None:
        return err
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'mid'), '%d' % wmarks.mid_permil)
    if err is not None:
        return err
    return _damo_fs.write_file(
            os.path.join(dir_path, 'low'), '%d' % wmarks.low_permil)

def write_quota_goal_dir(dir_path, goal):
    # goal metric is wip as of 6.8-rc4 days.
    if (not os.path.isfile(os.path.join(dir_path, 'target_metric')) and
        goal.metric != 'user_input'):
        return 'the kernel is not supporting quota goal metric'

    if os.path.isfile(os.path.join(dir_path, 'target_metric')):
        err = _damo_fs.write_file(os.path.join(dir_path, 'target_metric'),
                                  '%s' % goal.metric)
        if err is not None:
            return err

    if goal.has_nid():
        err = _damo_fs.write_file(os.path.join(dir_path, 'nid'), '%d' % goal.nid)
        if err is not None:
            return err

    if goal.has_memcg_path():
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'path'), '%s' % goal.memcg_path)
        if err is not None:
            return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'target_value'),
            '%d' % goal.target_value)
    if err is not None:
        return err

    return _damo_fs.write_file(
            os.path.join(dir_path, 'current_value'),
            '%d' % goal.current_value)

def write_quota_goals_dir(dir_path, goals):
    # goals dir has merged in 6.8-rc1
    if not os.path.isdir(dir_path):
        if len(goals) == 0:
            return None
        return 'the kernel is not supporting schemes quota goals'

    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_goals'), goals)
    if err is not None:
        return err

    for idx, goal in enumerate(goals):
        err = write_quota_goal_dir(os.path.join(dir_path, '%d' % idx), goal)
        if err is not None:
            return err
    return None

def write_quota_weights_dir(dir_path, quotas):
    err = _damo_fs.write_file(os.path.join(dir_path, 'sz_permil'),
                              '%d' % quotas.weight_sz_permil)
    if err is not None:
        return err

    err = _damo_fs.write_file(os.path.join(dir_path, 'nr_accesses_permil'),
                              '%d' % quotas.weight_nr_accesses_permil)
    if err is not None:
        return err

    return _damo_fs.write_file(os.path.join(dir_path, 'age_permil'),
                              '%d' % quotas.weight_age_permil)

def write_quota_fail_charge_ratio_files(dir_path, quotas):
    num_file = os.path.join(dir_path, 'fail_charge_num')
    if not os.path.isfile(num_file):
        return None
    err = _damo_fs.write_file(num_file, '%d' % quotas.fail_charge_num)
    if err is not None:
        return err

    denom_file = os.path.join(dir_path, 'fail_charge_denom')
    return _damo_fs.write_file(denom_file, '%d' % quotas.fail_charge_denom)

def write_quotas_dir(dir_path, quotas):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'ms'), '%d' % quotas.time_ms)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'bytes'), '%d' % quotas.sz_bytes)
    if err is not None:
        return err

    err = _damo_fs.write_file(os.path.join(dir_path, 'reset_interval_ms'),
                              '%d' % quotas.reset_interval_ms)
    if err is not None:
        return err

    err = write_quota_weights_dir(os.path.join(dir_path, 'weights'), quotas)
    if err is not None:
        return err

    err = write_quota_fail_charge_ratio_files(dir_path, quotas)
    if err is not None:
        return err

    err = write_quota_goals_dir(os.path.join(dir_path, 'goals'), quotas.goals)
    if err is not None:
        return err

    goal_tuner_file = os.path.join(dir_path, 'goal_tuner')
    if os.path.isfile(goal_tuner_file):
        return _damo_fs.write_file(goal_tuner_file, quotas.goal_tuner)

def write_scheme_dests_dir(dir_path, dests):
    if len(dests) == 0:
        return None
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_dests'), dests)
    if err is not None:
        return err

    for idx, dest in enumerate(dests):
        err = _damo_fs.write_file(os.path.join(dir_path, '%d' % idx, 'id'),
                                  '%d' % dest.id)
        if err is not None:
            return err
        err = _damo_fs.write_file(os.path.join(dir_path, '%d' % idx, 'weight'),
                                  '%d' % dest.weight)
        if err is not None:
            return err

def write_ulong(file_path, val):
    return _damo_fs.write_file(
            file_path, '%d' % min(val, _damo_fmt_str.ulong_max))

def write_scheme_access_pattern_dir(dir_path, pattern):
    err = write_ulong(
            os.path.join(dir_path, 'sz', 'min'), pattern.sz_bytes[0])
    if err is not None:
        return err

    err = write_ulong(
            os.path.join(dir_path, 'sz', 'max'), pattern.sz_bytes[1])
    if err is not None:
        return err

    err = write_ulong(os.path.join(dir_path, 'nr_accesses', 'min'),
                      pattern.nr_acc_min_max[0].samples)
    if err is not None:
        return err

    err = write_ulong(os.path.join(dir_path, 'nr_accesses', 'max'),
                      pattern.nr_acc_min_max[1].samples)
    if err is not None:
        return err

    err = write_ulong(os.path.join(dir_path, 'age', 'min'),
                      pattern.age_min_max[0].aggr_intervals)
    if err is not None:
        return err
    return write_ulong(os.path.join(dir_path, 'age', 'max'),
                       pattern.age_min_max[1].aggr_intervals)

def write_scheme_dir(dir_path, scheme):
    err = write_scheme_access_pattern_dir(
            os.path.join(dir_path, 'access_pattern'), scheme.access_pattern)
    if err is not None:
        return err

    err = _damo_fs.write_file(os.path.join(dir_path, 'action'), scheme.action)
    if err is not None:
        return err

    err = write_scheme_dests_dir(
            os.path.join(dir_path, 'dests'), scheme.dests)
    if err is not None:
        return err

    err = write_quotas_dir(os.path.join(dir_path, 'quotas'), scheme.quotas)
    if err is not None:
        return err

    err = write_watermarks_dir(
            os.path.join(dir_path, 'watermarks'), scheme.watermarks)
    if err is not None:
        return err

    if scheme.target_nid is not None:
        err = _damo_fs.write_file(os.path.join(dir_path,'target_nid'), '%s' % scheme.target_nid)
        if err is not None:
            return err

    err = write_filters_dirs(dir_path, scheme.filters)
    if err is not None:
        return err

    apply_interval_file = os.path.join(dir_path, 'apply_interval_us')
    # schemes apply interval is merged in v6.7-rc1
    if os.path.isfile(apply_interval_file):
        err = _damo_fs.write_file(apply_interval_file,
                                  '%d' % scheme.apply_interval_us)
        if err is not None:
            return err
    else:
        if scheme.apply_interval_us:
            return 'the kernel is not supporting schemes apply interval'

    max_nr_snapshots_file = os.path.join(dir_path, 'stats', 'max_nr_snapshots')
    if os.path.isfile(max_nr_snapshots_file):
        err = _damo_fs.write_file(max_nr_snapshots_file,
                                  '%d' % scheme.stats.max_nr_snapshots)
        if err is not None:
            return err
    else:
        if scheme.stats.max_nr_snapshots:
            return 'the kernel is not supporting max_nr_snapshots'

    return None

def write_schemes_dir(dir_path, schemes):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_schemes'), schemes)
    if err is not None:
        return err

    for idx, scheme in enumerate(schemes):
        err = write_scheme_dir(os.path.join(dir_path, '%d' % idx), scheme)
        if err is not None:
            return err

def write_target_region_dir(dir_path, region):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'start'), '%d' % region.start)
    if err is not None:
        return err

    return _damo_fs.write_file(
            os.path.join(dir_path, 'end'), '%d' % region.end)

def write_target_regions_dir(dir_path, regions):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_regions'), regions)
    if err is not None:
        return err

    for idx, region in enumerate(regions):
        err = write_target_region_dir(
                os.path.join(dir_path, '%d' % idx), region)
        if err is not None:
            return err
    return None

def write_target_dir(dir_path, target):
    if target.pid is not None:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'pid_target'), '%s' % target.pid)
        if err is not None:
            return err

    obsolete_file = os.path.join(dir_path, 'obsolete_target')
    if os.path.isfile(obsolete_file):
        err = _damo_fs.write_file(
                obsolete_file, '1' if target.obsolete else '0')
        if err is not None:
            return err
    elif target.obsolete:
        return 'obsolete_target unsupported'

    return write_target_regions_dir(
            os.path.join(dir_path, 'regions'), target.regions)


def write_targets_dir(dir_path, targets):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_targets'), targets)
    if err is not None:
        return err

    for idx, target in enumerate(targets):
        err = write_target_dir(os.path.join(dir_path, '%d' % idx), target)
        if err is not None:
            return err
    return None

def write_monitoring_intervals_goal_dir(dir_path, goal):
    err = _damo_fs.write_file(os.path.join(dir_path, 'access_bp'), '%d' %
                              goal.access_bp)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'aggrs'), '%d' %
                              goal.aggrs)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'min_sample_us'), '%d' %
                              goal.min_sample_us)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'max_sample_us'), '%d' %
                              goal.max_sample_us)
    if err is not None:
        return err

def write_ops_attrs_dir(dir_path, ops_attrs):
    if not os.path.isdir(dir_path):
        return None
    err = _damo_fs.write_file(os.path.join(dir_path, 'use_reports'),
                               'Y' if ops_attrs.use_reports else 'N')
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'write_only'),
                               'Y' if ops_attrs.write_only else 'N')
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'cpus'), ops_attrs.cpus)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'tids'), ops_attrs.tids)
    if err is not None:
        return err

def write_probe_filter_dir(dir_path, filter):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'type'), filter.filter_type)
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'matching'),
                              'Y' if filter.matching else 'N')
    if err is not None:
        return err
    err = _damo_fs.write_file(os.path.join(dir_path, 'allow'),
                              'Y' if filter.allow else 'N')
    if err is not None:
        return err
    if filter.filter_type == 'memcg':
        err = _damo_fs.write_file(os.path.join(dir_path, 'path'), filter.path)
        if err is not None:
            return err
    return None

def write_probe_dir(dir_path, probe):
    filters_dir = os.path.join(dir_path, 'filters')
    err = ensure_nr_file_for(os.path.join(filters_dir, 'nr_filters'),
                             probe.filters)
    if err is not None:
        return err

    for idx, filter in enumerate(probe.filters):
        err = write_probe_filter_dir(
                os.path.join(filters_dir, '%d' % idx), filter)
        if err is not None:
            return err
    return None

def write_probes_dir(dir_path, probes):
    if not os.path.isdir(dir_path):
        return None
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_probes'), probes)
    if err is not None:
        return err

    for idx, probe in enumerate(probes):
        err = write_probe_dir(os.path.join(dir_path, '%d' % idx), probe)
        if err is not None:
            return err
    return None

def write_sample_filter_dir(dir_path, sample_filter):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'type'), sample_filter.filter_type)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'matching'),
            'Y' if sample_filter.matching else 'N')
    if err is not None:
        return err
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'allow'),
            'Y' if sample_filter.allow else 'N')
    if err is not None:
        return err

    if sample_filter.filter_type == _damon.damon_filter_type_cpumask:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'cpumask'), sample_filter.cpumask)
        if err is not None:
            return err
    elif sample_filter.filter_type == _damon.damon_filter_type_threads:
        err = _damo_fs.write_file(
                os.path.join(dir_path, 'tid_arr'), sample_filter.tid_arr)
        if err is not None:
            return err
    return None

def write_sample_filters_dir(dir_path, filters):
    if not os.path.isdir(dir_path):
        if len(filters) == 0:
            return None
        return 'the kernel is not supporting sample filters'

    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_filters'), filters)
    if err is not None:
        return err

    for idx, sample_filter in enumerate(filters):
        err = write_sample_filter_dir(
                os.path.join(dir_path, '%d' % idx), sample_filter)
        if err is not None:
            return err
    return None

    err = _damo_fs.write_file

def write_sample_control_dir(dir_path, sample_control):
    if not os.path.isdir(dir_path):
        return None
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'primitives', 'page_table'),
            'Y' if sample_control.primitives_enabled.page_table else 'N')
    if err is not None:
        return err
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'primitives', 'page_fault'),
            'Y' if sample_control.primitives_enabled.page_fault else 'N')
    if err is not None:
        return err
    return write_sample_filters_dir(
            os.path.join(dir_path, 'filters'), sample_control.sample_filters)

def write_monitoring_attrs_dir(dir_path, context):
    err = _damo_fs.write_file(
            os.path.join(dir_path, 'intervals', 'sample_us'),
            '%d' % context.intervals.sample)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'intervals', 'aggr_us'),
            '%d' % context.intervals.aggr)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'intervals', 'update_us'),
            '%d' % context.intervals.ops_update)
    if err is not None:
        return err

    intervals_goal_path = os.path.join(dir_path, 'intervals', 'intervals_goal')
    if os.path.isdir(intervals_goal_path):
        err = write_monitoring_intervals_goal_dir(
                intervals_goal_path, context.intervals.intervals_goal)
        if err is not None:
            return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'nr_regions', 'min'),
            '%d' % context.nr_regions.minimum)
    if err is not None:
        return err

    err = _damo_fs.write_file(
            os.path.join(dir_path, 'nr_regions', 'max'),
            '%d' % context.nr_regions.maximum)
    if err is not None:
        return err

    err = write_probes_dir(
            os.path.join(dir_path, 'probes'), context.probes)
    if err is not None:
        return err

    return write_sample_control_dir(
            os.path.join(dir_path, 'sample'), context.sample_control)

def write_pause(pause_file, pause_val):
    if not os.path.isfile(pause_file):
        return None
    return _damo_fs.write_file(pause_file, 'Y' if pause_val else 'N')

def write_addr_unit(addr_unit_file, addr_unit_val):
    if not os.path.isfile(addr_unit_file):
        return None
    return _damo_fs.write_file(addr_unit_file, '%d' % addr_unit_val)

def write_context_dir(dir_path, context):
    err = _damo_fs.write_file(os.path.join(dir_path, 'operations'),
                              context.ops)
    if err is not None:
        return err

    err = write_ops_attrs_dir(
            os.path.join(dir_path, 'operations_attrs'), context.ops_attrs)
    if err is not None:
        return err

    err = write_monitoring_attrs_dir(
            os.path.join(dir_path, 'monitoring_attrs'), context)
    if err is not None:
        return err

    err = write_pause(os.path.join(dir_path, 'pause'), context.pause)
    if err is not None:
        return err

    err = write_addr_unit(os.path.join(dir_path, 'addr_unit'),
                          context.addr_unit)
    if err is not None:
        return err

    err = write_targets_dir(
            os.path.join(dir_path, 'targets'), context.targets)
    if err is not None:
        return err

    for scheme in context.schemes:
        scheme.access_pattern = scheme.access_pattern.converted_for_units(
                _damon.unit_samples, _damon.unit_aggr_intervals,
                context.intervals)
    return write_schemes_dir(
            os.path.join(dir_path, 'schemes'), context.schemes)

def write_contexts_dir(dir_path, contexts):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_contexts'), contexts)
    if err is not None:
        return err

    for idx, context in enumerate(contexts):
        err = write_context_dir(
                os.path.join(dir_path, '%d' % idx), context)
        if err is not None:
            return err

def write_kdamond_dir(dir_path, kdamond):
    refresh_ms_path = os.path.join(dir_path, 'refresh_ms')
    if os.path.isfile(refresh_ms_path):
        err = _damo_fs.write_file(refresh_ms_path, '%d' % kdamond.refresh_ms)
        if err is not None:
            return err

    err = write_contexts_dir(
            os.path.join(dir_path, 'contexts'), kdamond.contexts)
    if err is not None:
        return err

def write_kdamonds_dir(dir_path, kdamonds):
    err = ensure_nr_file_for(os.path.join(dir_path, 'nr_kdamonds'), kdamonds)
    if err:
        return err

    for idx, kdamond in enumerate(kdamonds):
        err = write_kdamond_dir(os.path.join(dir_path, '%d' % idx), kdamond)
        if err is not None:
            return err

def stage_kdamonds(kdamonds):
    """Write DAMON parameters for kdamonds to the sysfs files.

    Args:
        kdamonds: A list of _damon.Kdamond objects.

    Returns:
        None for success, an error string if failed.
    """
    # Assume caller checked supported()
    return write_kdamonds_dir(get_kdamonds_dir(), kdamonds)

def stage_kdamonds_targets(kdamonds):
    kdamonds_dir_path = get_kdamonds_dir()

    err = ensure_nr_file_for(os.path.join(kdamonds_dir_path, 'nr_kdamonds'), kdamonds)
    if err:
        return err

    for k_idx, kdamond in enumerate(kdamonds):
        contexts_dir_path = os.path.join(kdamonds_dir_path, '%d' % k_idx, 'contexts')
        contexts = kdamond.contexts

        err = ensure_nr_file_for(os.path.join(contexts_dir_path, 'nr_contexts'), contexts)
        if err:
            return err

        for c_idx, context in enumerate(contexts):
            targets_dir_path = os.path.join(contexts_dir_path, '%d' % c_idx, 'targets')
            err = write_targets_dir(targets_dir_path, context.targets)
            if err:
                return err

    return None

# for current_kdamonds()

def numbered_dirs_content(files_content, nr_filename):
    nr_dirs = int(files_content[nr_filename])
    number_dirs = []
    for i in range(nr_dirs):
        number_dirs.append(files_content['%d' % i])
    return number_dirs

def number_sorted_dirs(files_content):
    number_dirs = {}
    for filename, content in files_content.items():
        try:
            nr = int(filename)
        except:
            continue
        if type(content) != dict:
            continue
        number_dirs[nr] = content
    sorted_numbers = sorted(number_dirs.keys())
    return [number_dirs[nr] for nr in sorted_numbers]

def files_content_to_access_pattern(files_content):
    return _damon.DamosAccessPattern(
            [int(files_content['sz']['min']),
                int(files_content['sz']['max'])],
            [int(files_content['nr_accesses']['min']),
                int(files_content['nr_accesses']['max'])],
            _damon.unit_samples, # nr_accesses_unit
            [int(files_content['age']['min']),
                int(files_content['age']['max'])],
            _damon.unit_aggr_intervals) # age_unit

def files_content_to_quota_goals(files_content):
    goals = []
    for goal_kv in number_sorted_dirs(files_content):
        if 'target_metric' in goal_kv:
            if 'path' in goal_kv:
                memcg_path = goal_kv['path'].strip()
            else:
                memcg_path = None
            goals.append(
                    _damon.DamosQuotaGoal(
                        metric=goal_kv['target_metric'].strip(),
                        nid=goal_kv['nid'] if 'nid' in goal_kv else None,
                        memcg_path=memcg_path,
                        target_value=goal_kv['target_value'],
                        current_value=goal_kv['current_value']))
        else:
            goals.append(
                    _damon.DamosQuotaGoal(
                        target_value=goal_kv['target_value'],
                        current_value=goal_kv['current_value']))
    return goals

def files_content_to_quotas(files_content):
    fail_charge_num = files_content.get('fail_charge_num', 0)
    fail_charge_denom = files_content.get('fail_charge_denom', 0)
    return _damon.DamosQuotas(
            int(files_content['ms']),
            int(files_content['bytes']),
            int(files_content['reset_interval_ms']),
            [int(files_content['weights']['sz_permil']),
                int(files_content['weights']['nr_accesses_permil']),
                int(files_content['weights']['age_permil'])],
            files_content_to_quota_goals(files_content['goals'])
            if 'goals' in files_content else [],
            goal_tuner=files_content.get('goal_tuner', None),
            effective_sz_bytes=int(files_content['effective_bytes'])
            if 'effective_bytes' in files_content else 0,
            fail_charge_num=fail_charge_num,
            fail_charge_denom=fail_charge_denom)

def files_content_to_watermarks(files_content):
    return _damon.DamosWatermarks(
            files_content['metric'].strip(),
            int(files_content['interval_us']),
            int(files_content['high']),
            int(files_content['mid']),
            int(files_content['low']))

def files_content_to_damos_filter(files_content):
    allow = False
    if 'allow' in files_content:
        allow = files_content['allow'].strip()
    # the file name has changed from pass to allow after v1 patch
    elif 'pass' in files_content:
        allow = files_content['pass'].strip()

    hugepage_size=None
    if 'min' in files_content and 'max' in files_content:
        hugepage_size = [_damo_fmt_str.text_to_bytes(x)
                         for x in [files_content['min'], files_content['max']]]

    return _damon.DamosFilter(
        files_content['type'].strip(),
        files_content['matching'].strip(),
        allow,
        files_content['memcg_path'].strip(),
        _damon.DamonRegion(files_content['addr_start'].strip(),
                           files_content['addr_end'].strip())
        if 'addr_start' in files_content and 'addr_end' in files_content
        else None,
        files_content['damon_target_idx']
        if 'damon_target_idx' in files_content else None,
        hugepage_size=hugepage_size)

def files_content_to_damos_filters(scheme_files_content):
    filters = []
    if 'core_filters' in scheme_files_content:
        filters += [files_content_to_damos_filter(filter_kv)
                    for filter_kv in numbered_dirs_content(
                        scheme_files_content['core_filters'], 'nr_filters')]
        filters += [files_content_to_damos_filter(filter_kv)
                    for filter_kv in numbered_dirs_content(
                        scheme_files_content['ops_filters'], 'nr_filters')]
    filters += [files_content_to_damos_filter(filter_kv)
                for filter_kv in numbered_dirs_content(
                    scheme_files_content['filters'], 'nr_filters')]
    return filters

def files_content_to_damos_stats(files_content):
    if 'nr_snapshots' in files_content:
        nr_snapshots = int(files_content['nr_snapshots'])
    else:
        nr_snapshots = 0
    if 'max_nr_snapshots' in files_content:
        max_nr_snapshots = int(files_content['max_nr_snapshots'])
    else:
        max_nr_snapshots = 0
    return _damon.DamosStats(
            int(files_content['nr_tried']),
            int(files_content['sz_tried']),
            int(files_content['nr_applied']),
            int(files_content['sz_applied']),
            int(files_content['sz_ops_filter_passed']
                if 'sz_ops_filter_passed' in files_content else 0),
            int(files_content['qt_exceeds']),
            nr_snapshots=nr_snapshots, max_nr_snapshots=max_nr_snapshots)

def files_content_to_probe_hits(files_content):
    hits = []
    for kv in number_sorted_dirs(files_content):
        hits.append(int(kv['hits']))
    return hits

def files_content_to_damos_tried_regions(files_content):
    regions = []
    for kv in number_sorted_dirs(files_content):
        if 'probes' in kv:
            probe_hits = files_content_to_probe_hits(kv['probes'])
        else:
            probe_hits = []
        regions.append(
                _damon.DamonRegion(
                    int(kv['start']), int(kv['end']),
                    int(kv['nr_accesses']), _damon.unit_samples,
                    int(kv['age']), _damon.unit_aggr_intervals,
                    int(kv['sz_filter_passed'])
                    if 'sz_filter_passed' in kv else None,
                    probe_hits=probe_hits))
    return regions

def files_content_to_damos_dest(files_content):
    return _damon.DamosDest(
            id=int(files_content['id']), weight=int(files_content['weight']))

def files_content_to_scheme(files_content):
    dests = []
    if 'dests' in files_content:
        dests = [files_content_to_damos_dest(content)
            for content in numbered_dirs_content(
                files_content['dests'], 'nr_dests')]
    return _damon.Damos(
            files_content_to_access_pattern(files_content['access_pattern']),
            files_content['action'].strip(),
            files_content['target_nid'].strip()
                if 'target_nid' in files_content else None,
            files_content['apply_interval_us'].strip()
                if 'apply_interval_us' in files_content else None,
            files_content_to_quotas(files_content['quotas']),
            files_content_to_watermarks(files_content['watermarks']),
            files_content_to_damos_filters(files_content)
                if 'filters' in files_content else [],
            files_content_to_damos_stats(files_content['stats']),
            files_content_to_damos_tried_regions(
                files_content['tried_regions'])
                if 'tried_regions' in files_content else [],
            files_content['tried_regions']['total_bytes']
                if 'tried_regions' in files_content and
                    'total_bytes' in files_content['tried_regions'] else None,
            dests=dests)

def files_content_to_regions(files_content):
    return [_damon.DamonRegion(
            int(kv['start']), int(kv['end']))
            for kv in numbered_dirs_content(files_content, 'nr_regions')]

def files_content_to_target(files_content):
    try:
        pid = int(files_content['pid_target'])
    except:
        pid = None
    obsolete = False
    if 'obsolete_target' in files_content:
        obsolete = files_content['obsolete_target'].strip()
    regions = files_content_to_regions(files_content['regions'])
    return _damon.DamonTarget(pid, regions, obsolete=obsolete)

def files_content_to_damon_filter(files_content):
    path = None
    if 'path' in files_content:
        files_content.get('path').strip()
    return _damon.DamonFilter(
        filter_type=files_content['type'].strip(),
        matching=files_content['matching'].strip(),
        allow=files_content['allow'].strip(), path=path)

def files_content_to_probe(files_content):
    filters_content = files_content['filters']
    filters = []
    for kv in number_sorted_dirs(filters_content):
        filter = files_content_to_damon_filter(kv)
        filters.append(filter)
    return _damon.DamonProbe(filters=filters)

def files_content_to_probes(files_content):
    probes = []
    for kv in number_sorted_dirs(files_content):
        probes.append(files_content_to_probe(kv))
    return probes

def files_content_to_sample_filter(files_content):
    filter_type = files_content['type'].strip()
    matching = files_content['matching'].strip()
    allow = files_content['matching'].strip()
    cpumask = files_content['cpumask'].strip()
    tid_arr = files_content['tid_arr'].strip()
    return _damon.DamonSampleFilter(
            filter_type=filter_type, matching=matching, allow=allow,
            cpumask=cpumask, tid_arr=tid_arr)

def files_content_to_sample_filters(files_content):
    return [files_content_to_sample_filter(filter_kv)
                for filter_kv in numbered_dirs_content(
                    files_content, 'nr_filters')]

def files_content_to_sample_control(files_content):
    page_table = files_content['primitives']['page_table'].strip()
    page_fault = files_content['primitives']['page_fault'].strip()
    primitives_enabled = _damon.DamonPrimitivesEnabled(
            page_table=page_table, page_fault=page_fault)
    sample_filters = files_content_to_sample_filters(files_content['filters'])
    return _damon.DamonSampleControl(
            primitives_enabled=primitives_enabled,
            sample_filters=sample_filters)

def files_content_to_ops_attrs(files_content):
    use_reports = files_content['use_reports'].strip()
    write_only = files_content['write_only'].strip()
    cpus = files_content['cpus'].strip()
    tids = files_content['tids'].strip()
    return _damon.OpsAttrs(use_reports=use_reports, write_only=write_only,
                           cpus=cpus, tids=tids)

def files_content_to_context(files_content):
    mon_attrs_content = files_content['monitoring_attrs']
    intervals_content = mon_attrs_content['intervals']
    if 'intervals_goal' in intervals_content:
        kvpairs = intervals_content['intervals_goal']
        intervals_goal = _damon.DamonIntervalsGoal(
                int(kvpairs['access_bp']), int(kvpairs['aggrs']),
                int(kvpairs['min_sample_us']), int(kvpairs['max_sample_us']))
    else:
        intervals_goal = _damon.DamonIntervalsGoal()
    intervals = _damon.DamonIntervals(
            int(intervals_content['sample_us']),
            int(intervals_content['aggr_us']),
            int(intervals_content['update_us']),
            intervals_goal)
    nr_regions_content = mon_attrs_content['nr_regions']
    nr_regions = _damon.DamonNrRegionsRange(
            int(nr_regions_content['min']),
            int(nr_regions_content['max']))
    if 'probes' in files_content['monitoring_attrs']:
        probes = files_content_to_probes(
                files_content['monitoring_attrs']['probes'])
    else:
        probes = []
    if 'sample' in mon_attrs_content:
        sample_control = files_content_to_sample_control(
                mon_attrs_content['sample'])
    else:
        sample_control = _damon.DamonSampleControl()
    ops = files_content['operations'].strip()
    if 'operations_attrs' in files_content:
        ops_attrs = files_content_to_ops_attrs(
                files_content['operations_attrs'])
    else:
        ops_attrs = _damon.OpsAttrs()
    if 'pause' in files_content:
        pause = files_content['pause'].strip()
    else:
        pause = False
    addr_unit = files_content.get('addr_unit', 1)

    targets_content = files_content['targets']
    targets = [files_content_to_target(content)
            for content in numbered_dirs_content(
                targets_content, 'nr_targets')]

    schemes_content = files_content['schemes']
    schemes = [files_content_to_scheme(content)
            for content in numbered_dirs_content(
                schemes_content, 'nr_schemes')]

    return _damon.DamonCtx(ops, targets, intervals, nr_regions, schemes,
                           ops_attrs=ops_attrs, sample_control=sample_control,
                           pause=pause, addr_unit=addr_unit, probes=probes)

def files_content_to_kdamond(files_content):
    contexts_content = files_content['contexts']
    contexts = [files_content_to_context(content)
            for content in numbered_dirs_content(
                contexts_content, 'nr_contexts')]
    state = files_content['state'].strip()
    pid = files_content['pid'].strip()
    if 'refresh_ms' in files_content:
        refresh_ms = files_content['refresh_ms'].strip()
    else:
        refresh_ms = 0
    return _damon.Kdamond(state, pid, refresh_ms=refresh_ms, contexts=contexts)

def files_content_to_kdamonds(files_contents):
    return [files_content_to_kdamond(content)
            for content in numbered_dirs_content(
                files_contents, 'nr_kdamonds')]

def current_kdamonds():
    # Assume caller checked supported()
    return files_content_to_kdamonds(_damo_fs.read_files(get_kdamonds_dir()))

def get_nr_kdamonds_file():
    return os.path.join(get_kdamonds_dir(), 'nr_kdamonds')

def nr_kdamonds():
    nr_kdamonds, err = _damo_fs.read_file(get_nr_kdamonds_file())
    if err != None:
        raise Exception('nr_kdamonds_file read fail (%s)' % err)
    return int(nr_kdamonds)

def commit_staged(kdamond_idxs):
    for kdamond_idx in kdamond_idxs:
        err = _damo_fs.write_file(get_state_file_of(kdamond_idx), 'commit')
        if err != None:
            return err
    return None

def commit_quota_goals(kdamond_idxs):
    for kdamond_idx in kdamond_idxs:
        err = _damo_fs.write_file(get_state_file_of(kdamond_idx),
                'commit_schemes_quota_goals')
        if err != None:
            return err

# features

# sysfs was merged in v5.18-rc1
features_sysfs_support_from_begining = [
        'sysfs/schemes',
        'sysfs/init_regions',
        'sysfs/vaddr',
        'sysfs/paddr',
        'sysfs/schemes_size_quota',
        'sysfs/schemes_time_quota',
        'sysfs/schemes_prioritization',
        'sysfs/schemes_wmarks',
        'sysfs/schemes_stat_succ',
        'sysfs/schemes_stat_qt_exceed',
        ]

def kdamond_dir_of(kdamond_idx):
    return os.path.join(get_kdamonds_dir(), '%s' % kdamond_idx)

def ctx_dir_of(kdamond_idx, context_idx):
    return os.path.join(
            kdamond_dir_of(kdamond_idx), 'contexts', '%s' % context_idx)

def target_dir_of(kdamond_idx, context_idx, target_idx):
    return os.path.join(
            ctx_dir_of(kdamond_idx, context_idx), 'targets',
            '%s' % target_idx)

def schemes_dir_of(kdamond_idx, context_idx):
    return os.path.join(ctx_dir_of(kdamond_idx, context_idx), 'schemes')

def _avail_ops():
    '''Assumes called by update_supported_features() assuming one scheme.
    Returns available ops input and error'''
    avail_ops = []
    avail_operations_filepath = os.path.join(ctx_dir_of(0, 0),
            'avail_operations')
    if not os.path.isfile(avail_operations_filepath):
        operations_filepath = os.path.join(ctx_dir_of(0, 0), 'operations')
        for ops in ['vaddr', 'paddr', 'fvaddr']:
            err = _damo_fs.write_file(operations_filepath, ops)
            if err != None:
                avail_ops.append(ops)
        return avail_ops, None

    content, err = _damo_fs.read_file(avail_operations_filepath)
    if err != None:
        return None, err
    return content.strip().split(), None

def scheme_dir_of(kdamond_idx, context_idx, scheme_idx):
    return os.path.join(
            schemes_dir_of(kdamond_idx, context_idx), '%s' % scheme_idx)

def scheme_tried_regions_dir_of(kdamond_idx, context_idx, scheme_idx):
    return os.path.join(
            scheme_dir_of(kdamond_idx, context_idx, scheme_idx),
            'tried_regions')

def mk_feature_supports_map():
    '''
    Returns a map indicating list of supported and unsupported DAMON features,
    and an error if making the map failed.
    Keys of the map are names of DAMON features.
    Values are bool indicating whether the feature is supported.
    '''
    supports_map = {x.name: False for x in _damon_features.features_list}

    for feature in features_sysfs_support_from_begining:
        supports_map[feature] = True

    orig_kdamonds = current_kdamonds()
    # While DAMON is running, feature checking I/O can fail, corrupt something,
    # or make something complicated.  Just don't do that.
    for kd in orig_kdamonds:
        if kd.state == 'on':
            return None, 'DAMON is running'

    kdamonds_for_feature_check = [
            _damon.Kdamond(
                state=None, pid=None, contexts=[
                    _damon.DamonCtx(
                        targets=[_damon.DamonTarget(
                            pid=None, regions=[])],
                        schemes=[_damon.Damos()])])]
    err = stage_kdamonds(kdamonds_for_feature_check)
    if err is not None:
        stage_kdamonds(orig_kdamonds)
        return None, 'staging feature check purpose kdamond failed'

    if os.path.isdir(scheme_tried_regions_dir_of(0, 0, 0)):
        supports_map['sysfs/schemes_tried_regions'] = True

    if os.path.isfile(os.path.join(scheme_tried_regions_dir_of(0, 0, 0),
            'total_bytes')):
        supports_map['sysfs/schemes_tried_regions_sz'] = True
        # address and target filter types are added in v6.6-rc1, together with
        # schemes_tried_regions_sz
        supports_map['sysfs/schemes_filters_addr'] = True
        supports_map['sysfs/schemes_filters_target'] = True

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'filters')):
        supports_map['sysfs/schemes_filters'] = True
        # anon and memcg were supported from the beginning
        supports_map['sysfs/schemes_filters_anon'] = True
        supports_map['sysfs/schemes_filters_memcg'] = True
        kdamonds_for_feature_check[0].contexts[0].schemes[0].filters = [
                _damon.DamosFilter('young', True)]
        err = stage_kdamonds(kdamonds_for_feature_check)
        if err is None:
            supports_map['sysfs/schemes_filters_young'] = True

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'apply_interval_us')):
        supports_map['sysfs/schemes_apply_interval'] = True

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals')):
        supports_map['sysfs/schemes_quota_goals'] = True

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'quotas',
                                   'effective_bytes')):
        supports_map['sysfs/schemes_quota_effective_bytes'] = True
        # goal_metric and goal_some_psi will be merged together with effective bytes.
        supports_map['sysfs/schemes_quota_goal_metric'] = True
        supports_map['sysfs/schemes_quota_goal_some_psi'] = True

    if os.path.isfile(os.path.join(scheme_dir_of(0, 0, 0), 'target_nid')):
        supports_map['sysfs/schemes_migrate'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0),
                         'stats', 'sz_ops_filter_passed')):
        supports_map['sysfs/sz_ops_filter_passed'] = True

    ops_filters_dir = os.path.join(scheme_dir_of(0, 0, 0), 'ops_filters')
    if not os.path.isdir(ops_filters_dir):
        ops_filters_dir = os.path.join(scheme_dir_of(0, 0, 0), 'filters')

    if os.path.isfile(
            os.path.join(ops_filters_dir, '0', 'allow')):
        supports_map['sysfs/allow_filter'] = True

    if os.path.isfile(
            os.path.join(ops_filters_dir, '0', 'min')):
        supports_map['sysfs/schemes_filters_hugepage_size'] = True

    if os.path.isdir(
            os.path.join(ctx_dir_of(0, 0),
                         'monitoring_attrs', 'intervals', 'intervals_goal')):
        supports_map['sysfs/intervals_goal'] = True

    if os.path.isdir(
            os.path.join(scheme_dir_of(0, 0, 0), 'core_filters')):
        supports_map['sysfs/schemes_filters_core_ops_dirs'] = True

        # unmapped and active pages DAMOS filters are merged into v6.15
        # together with core_ops_dirs
        supports_map['sysfs/schemes_filters_unmapped'] = True
        supports_map['sysfs/schemes_filters_active'] = True

    if supports_map['sysfs/schemes_quota_goals'] is True:
        kdamonds_for_feature_check = [
                _damon.Kdamond(
                    state=None, pid=None, contexts=[
                        _damon.DamonCtx(
                            targets=[_damon.DamonTarget(
                                pid=None, regions=[])],
                            schemes=[_damon.Damos(
                                quotas=_damon.DamosQuotas(
                                    goals=[_damon.DamosQuotaGoal()])
                                )])])]
        err = stage_kdamonds(kdamonds_for_feature_check)
        if err is not None:
            stage_kdamonds(orig_kdamonds)
            return None, \
                    'staging damos goal feature check purpose kdamond failed'

        if os.path.isfile(
                os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals', '0',
                             'nid')):
            supports_map['sysfs/schemes_quota_goal_node_mem_used_free'] = True

        if os.path.isfile(
                os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goals', '0',
                             'path')):
            supports_map['sysfs/schemes_quota_goal_node_memcg_used_free'] = True

    if os.path.isdir(os.path.join(scheme_dir_of(0, 0, 0), 'dests')):
        supports_map['sysfs/schemes_dests'] = True

    if os.path.isfile(os.path.join(kdamond_dir_of(0), 'refresh_ms')):
        supports_map['sysfs/refresh_ms'] = True

    if os.path.isfile(os.path.join(ctx_dir_of(0, 0), 'addr_unit')):
        supports_map['sysfs/addr_unit'] = True

    if os.path.isfile(os.path.join(target_dir_of(0, 0, 0), 'obsolete_target')):
        supports_map['sysfs/obsolete_target'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0),
                         'stats', 'nr_snapshots')):
        supports_map['sysfs/damos_stat_nr_snapshots'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0),
                         'stats', 'max_nr_snapshots')):
        supports_map['sysfs/damos_max_nr_snapshots'] = True
        # will be upstreamed together with max_nr_snapshots
        supports_map['sysfs/damos_quota_goal_in_active_mem_bp'] = True

    if os.path.isdir(
            os.path.join(ctx_dir_of(0, 0),
                         'monitoring_attrs', 'sample')):
                supports_map['sysfs/damon_sample_control'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0), 'quotas', 'goal_tuner')):
        supports_map['sysfs/damos_quota_goal_tuner'] = True

    if os.path.isfile(os.path.join(ctx_dir_of(0, 0), 'pause')):
        supports_map['sysfs/ctx_pause'] = True

    if os.path.isfile(
            os.path.join(scheme_dir_of(0, 0, 0), 'quotas',
                         'fail_charge_denom')):
        supports_map['sysfs/damos_quota_fail_charge_ratio'] = True
        # damos_collapse and damos_node_eligible_mem_bp will be upstreamed
        # together.
        supports_map['sysfs/damos_action_collapse'] = True
        supports_map['sysfs/damos_quota_goal_node_eligible_mem_bp'] = True

    if os.path.isdir(
            os.path.join(ctx_dir_of(0, 0), 'monitoring_attrs', 'probes')):
        supports_map['sysfs/attrs_monitoring'] = True

    if os.path.isdir(os.path.join(ctx_dir_of(0, 0), 'operations_attrs')):
        supports_map['sysfs/ops_attrs'] = True

    avail_operations_filepath = os.path.join(ctx_dir_of(0, 0),
            'avail_operations')
    if os.path.isfile(avail_operations_filepath):
        supports_map['sysfs/avail_ops'] = True
        supports_map['sysfs/online_params_commit'] = True

    avail_ops, err = _avail_ops()
    if err == None:
        for ops in ['vaddr', 'paddr', 'fvaddr']:
            supports_map['sysfs/%s' % ops] = ops in avail_ops
    err = stage_kdamonds(orig_kdamonds)
    if err is not None:
        return None, 'restoring original kdamonds setup failed'
    return supports_map, None



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import unittest

import _test_damo_common

_test_damo_common.add_damo_dir_to_syspath()

import _damo_records
import _damon

class TestDamon(unittest.TestCase):
    def test_parse_file_permission_str(self):
        perm, err = _damo_records.parse_file_permission_str('777')
        self.assertEqual(perm, 0o777)
        self.assertIsNone(err)

        perm, err = _damo_records.parse_file_permission_str('-123')
        self.assertIsNone(perm)
        self.assertIsNotNone(err)

        perm, err = _damo_records.parse_file_permission_str('1000')
        self.assertIsNone(perm)
        self.assertIsNotNone(err)

        perm, err = _damo_records.parse_file_permission_str('778')
        self.assertIsNone(perm)
        self.assertIsNotNone(err)

    def test_record_from_kvpairs(self):
        records = [_damo_records.DamonRecord.from_kvpairs(p) for p in [
            {
                "kdamond_idx": 0, "context_idx": 0,
                "intervals": {
                    "sample_us": "5ms", "aggr_us": "100ms",
                    "ops_update_us": "1s"
                    },
                "scheme_idx": 0, "target_id": None,
                "snapshots": [
                    {
                        "start_time": "1690134099856513792",
                        "end_time": "1690134099956513792",
                        "regions": [
                            {
                                "start": "4,294,967,296",
                                "end": "7,516,192,768",
                                "nr_accesses": {
                                    "samples": 0,
                                    "percent": None
                                    },
                                "age": {
                                    "usec": None,
                                    "aggr_intervals": "5,913"
                                    }
                                }
                            ],
                        "total_bytes": "3221225472"
                        }
                    ]
                }
            ]]
        self.assertEqual(len(records), 1)
        self.assertEqual(len(records[0].snapshots), 1)

        # test DamonSnapshot.total_bytes added by commit 98c1bd8f467f6.  Test
        # older output that doesn't have the data.
        records = [_damo_records.DamonRecord.from_kvpairs(p) for p in [
            {
                "kdamond_idx": 0, "context_idx": 0,
                "intervals": {
                    "sample_us": "5ms", "aggr_us": "100ms",
                    "ops_update_us": "1s"
                    },
                "scheme_idx": 0, "target_id": None,
                "snapshots": [
                    {
                        "start_time": "1690134099856513792",
                        "end_time": "1690134099956513792",
                        "regions": [
                            {
                                "start": "4,294,967,296",
                                "end": "7,516,192,768",
                                "nr_accesses": {
                                    "samples": 0,
                                    "percent": None
                                    },
                                "age": {
                                    "usec": None,
                                    "aggr_intervals": "5,913"
                                    }
                                }
                            ],
                        }
                    ]
                }
            ]]
        self.assertEqual(len(records), 1)
        self.assertEqual(len(records[0].snapshots), 1)

    def test_filter_by_address(self):
        ranges = [[3, 5], [6, 9], [10, 12]]
        region = _damon.DamonRegion(7, 8)
        self.assertEqual(
                _damo_records.filter_by_addr(_damon.DamonRegion(7, 8),
                    ranges), [_damon.DamonRegion(7, 8)])
        self.assertEqual(
                _damo_records.filter_by_addr(_damon.DamonRegion(1, 2),
                    ranges), [])
        self.assertEqual(
                _damo_records.filter_by_addr(_damon.DamonRegion(5, 6),
                    ranges), [])
        self.assertEqual(
                _damo_records.filter_by_addr(_damon.DamonRegion(1, 4),
                    ranges), [_damon.DamonRegion(3, 4)])
        self.assertEqual(
                _damo_records.filter_by_addr(_damon.DamonRegion(1, 20),
                    ranges), [_damon.DamonRegion(3, 5), _damon.DamonRegion(6,
                        9), _damon.DamonRegion(10, 12)])

    def test_parse_sort_bytes_ranges_input(self):
        self.assertEqual(
                _damo_records.parse_sort_bytes_ranges_input([['1G', '2G']]),
                ([[1024 * 1024 * 1024, 1024 * 1024 * 1024 * 2]], None))
        ranges, err = _damo_records.parse_sort_bytes_ranges_input(
                [['abc', 'def']])
        self.assertNotEqual(err, None)
        ranges, err = _damo_records.parse_sort_bytes_ranges_input([[4, 3]])
        self.assertNotEqual(err, None)
        ranges, err = _damo_records.parse_sort_bytes_ranges_input(
                [[5, 7], [2, 6]])
        self.assertEqual(err, 'overlapping range')

        self.assertEqual(
                _damo_records.parse_sort_bytes_ranges_input(
                    [[10, 20], [5, 7]]), ([[5, 7], [10, 20]], None))

    def test_damon_trace_fields(self):
        # trace-cmd report output
        self.assertEqual(_damo_records.damon_trace_fields('cpus=8'), None)
        self.assertEqual(_damo_records.damon_trace_fields(
            '       kdamond.0-263515 [000] ..... 92566.021974: '
            'damon_aggregated:     target_id=0 nr_regions=11 '
            '4294967296-4701806592: 0 1635'),
            ['92566.021974:', 'damon_aggregated:', 'target_id=0',
             'nr_regions=11', '4294967296-4701806592:', '0', '1635'])
        self.assertEqual(_damo_records.damon_trace_fields(
            '       kdamond.0-263515 [000] ..... 92566.021974: '
            'damon_aggregated:     target_id=0 nr_regions=11 '
            '4294967296-4701806592: 0'),
            ['92566.021974:', 'damon_aggregated:', 'target_id=0',
             'nr_regions=11', '4294967296-4701806592:', '0'])

        # perf script output
        self.assertEqual(
                _damo_records.damon_trace_fields(
                    '        kthreadd  264573 [003] 93211.675342:'
                    '             damon:damon_aggregated: target_id=0 '
                    'nr_regions=2 4294967296-5518340096: 0'),
                '93211.675342: damon:damon_aggregated: target_id=0 nr_regions=2 '
                '4294967296-5518340096: 0'.split())

    def test_parse_damon_trace_region(self):
        # trace-cmd output
        self.assertEqual(_damo_records.parse_damon_trace_region(
            _damo_records.damon_trace_fields(
                '       kdamond.0-48034 [007] .....  5435.406849: '
                'damon_aggregated:     target_id=12 nr_regions=3 '
                '4294967296-8372879360: 4 5')),
            (_damon.DamonRegion(
                start=4294967296, end=8372879360, nr_accesses=4,
                nr_accesses_unit=_damon.unit_samples, age=5,
                age_unit=_damon.unit_aggr_intervals), 5435406849000, 12, 3))
        self.assertEqual(_damo_records.parse_damon_trace_region(
            _damo_records.damon_trace_fields(
                '       kdamond.0-48034 [007] .....  5435.406849: '
                'damon_aggregated:     target_id=12 nr_regions=3 '
                '4294967296-8372879360: 4')),
            (_damon.DamonRegion(
                start=4294967296, end=8372879360, nr_accesses=4,
                nr_accesses_unit=_damon.unit_samples, age=None,
                age_unit=_damon.unit_aggr_intervals), 5435406849000, 12, 3))

        # perf-script output
        self.assertEqual(_damo_records.parse_damon_trace_region(
            _damo_records.damon_trace_fields(
                '        kthreadd  264573 [002] 93214.744389:             '
                'damon:damon_aggregated: target_id=12 nr_regions=10 '
                '7473692672-8372879360: 3 4')),
            (_damon.DamonRegion(
                start=7473692672, end=8372879360, nr_accesses=3,
                nr_accesses_unit=_damon.unit_samples, age=4,
                age_unit=_damon.unit_aggr_intervals),
             93214744389000, 12, 10))
        self.assertEqual(_damo_records.parse_damon_trace_region(
            _damo_records.damon_trace_fields(
                '        kthreadd  264573 [002] 93214.744389:             '
                'damon:damon_aggregated: target_id=12 nr_regions=10 '
                '7473692672-8372879360: 3')),
            (_damon.DamonRegion(
                start=7473692672, end=8372879360, nr_accesses=3,
                nr_accesses_unit=_damon.unit_samples, age=None,
                age_unit=_damon.unit_aggr_intervals),
             93214744389000, 12, 10))

    def test_parse_damon_trace_intervals_tune(self):
        self.assertEqual(_damo_records.parse_damon_trace_intervals_tune(
            _damo_records.damon_trace_fields(
                '       kdamond.0-48034 [007] .....  5435.619246: '
                'damon_monitor_intervals_tune: sample_us=1234')),
            (True, 1234))
        self.assertEqual(_damo_records.parse_damon_trace_intervals_tune(
            _damo_records.damon_trace_fields(
                '       kdamond.0-48034 [007] .....  5435.406849: '
                'damon_aggregated:     target_id=12 nr_regions=3 '
                '4294967296-8372879360: 4 5')),
            (False, None))

        self.assertEqual(_damo_records.parse_damon_trace_intervals_tune(
            _damo_records.damon_trace_fields(
                '        kthreadd  264573 [003] 93211.464138: '
                'damon:damon_monitor_intervals_tune: sample_us=1234')),
            (True, 1234))
        self.assertEqual(_damo_records.parse_damon_trace_intervals_tune(
            _damo_records.damon_trace_fields(
                '        kthreadd  264573 [002] 93214.744389:             '
                'damon:damon_aggregated: target_id=12 nr_regions=10 '
                '7473692672-8372879360: 3')),
            (False, None))

if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: GPL-2.0

import datetime
import os
import sys
import tempfile

import _damon

def log(msg):
    msg = '%s: %s' % (datetime.datetime.now(), msg)
    with open('.damo_cli_complete_log', 'a') as f:
        f.write('%s\n' % msg)

completion_script = '''# SPDX-License-Identifier: GPL-2.0

# bash completion support for damo.
# To use this, 'source' this script.  For example,
#
#     $ source ./scripts/damo-completion.sh
#     $ ./damo [TAB][TAB]
#     help	record	report	start	stop	tune	version

_damo_complete()
{
	local cur prev words cword
	_init_completion || return

	if [ "$cword" -lt 1 ]
	then
		return 1
	fi

	_damo_complete_debug="false"
	if [ "${_damo_complete_debug}" = "true" ]
	then
		echo "cur '$cur'" >> .damo_completion_log
		echo "prev '$prev'" >> .damo_completion_log
		echo "words '${words[@]}'" >> .damo_completion_log
		echo "cword '$cword'" >> .damo_completion_log
		echo >> .damo_completion_log
	fi

	candidates=$("${words[0]}" --cli_complete "$cword" "${words[@]}")

	COMPREPLY=($(compgen -W "${candidates}" -- "$cur"))
	return 0
}

complete -F _damo_complete damo'''

class Option:
    name = None
    nr_args = None  # -1 for variable number of arguments
    repeatable = None

    # List of positional candidates for this option's arguments.
    # e.g., [['foo', 'bar'], None, ['baz', 'bow']] means for the first
    # argument, 'foo' or 'bar' can be entered.  No suggestion for the second
    # argument.  For third argument, 'baz' and 'bow' can be entered.
    positional_candidates = None

    # List of non-positional candidastes for this option's arguments.
    # e.g., ['foo', 'bar'] means for any argument for this option, 'foo' and
    # 'bar' can be suggested.
    non_positional_candidates = None

    def __init__(self, name, nr_args, repeatable=True,
                 positional_candidates=None,
                 non_positional_candidates=None):
        self.name = name
        self.nr_args = nr_args
        self.repeatable = repeatable
        self.positional_candidates = positional_candidates
        self.non_positional_candidates = non_positional_candidates

def prev_option_nr_filed_args(words, cword):
    for i in range(cword -1, -1, -1):
        if words[i].startswith('-'):
            prev_option = words[i]
            nr_filled_args = cword - 1 - i
            return prev_option, nr_filled_args
    return None, None

def can_suggest_options(words, cword, options):
    if cword == 0 or words[cword].startswith('-'):
        return True
    prev_option, nr_filled_args = prev_option_nr_filed_args(words, cword)
    if prev_option is None:
        return False
    for option in options:
        if option.name == prev_option and option.nr_args == nr_filled_args:
            return True
    return False

def option_arg_candidates(words, cword, options):
    prev_option, nr_filled_args = prev_option_nr_filed_args(words, cword)
    candidates = []
    for option in options:
        if option.name != prev_option:
            continue
        if option.non_positional_candidates is not None:
            candidates += option.non_positional_candidates
        if option.positional_candidates is None:
            continue
        if len(option.positional_candidates) > nr_filled_args:
            if option.positional_candidates[nr_filled_args] is not None:
                candidates += option.positional_candidates[nr_filled_args]
    return candidates

def get_candidates(words, cword, options):
    '''
    words and cword should start from the options part (no command).
    options should be a list of Option objects.
    '''
    if can_suggest_options(words, cword, options) is False:
        return option_arg_candidates(words, cword, options)
    candidates = []
    for option in options:
        if option.repeatable is False and option.name in words[:cword]:
            continue
        candidates.append(option.name)
    return candidates

damos_filter_types = ['active', 'memcg', 'young', 'hugepage_size', 'unmapped',
                      'addr', 'target']
damos_filter_positional_candids = [
        ['allow', 'reject'], ['none'] + damos_filter_types,
        # todo: do not suggest unless 'none' was entered
        damos_filter_types]

def damon_param_candidates(words, cword):
    return get_candidates(
            words, cword,
            [
                Option('--ops', 1, True, [['vaddr', 'paddr', 'fvaddr']]),
                Option('--pause_ctx', nr_args=-1, repeatable=False),
                Option('--monitoring_intervals_autotune', 0, False),
                Option('--numa_node', -1, True),
                Option('--monitoring_intervals', 3, True),
                Option('--monitoring_intervals_goal', 4, True),
                Option('--monitoring_nr_regions_range', 2, True),
                Option('--damos_action', 1, True,
                       [['willneed', 'cold', 'pageout', 'hugepage',
                         'nohugepage', 'lru_prio', 'lru_deprio', 'migrate_hot',
                         'migrate_cold', 'stat']]
                       ),
                Option('--damos_apply_interval', 1, True),
                Option('--damos_quota_interval', 1, True),
                Option('--damos_quota_space', 1, True),
                Option('--damos_quota_goal', -1, True, [_damon.qgoal_metrics]),
                Option(name='--damos_quota_goal_tuner', nr_args=1,
                       repeatable=True,
                       positional_candidates=[
                           [_damon.damos_qgtuner_consist,
                            _damon.damos_qgtuner_temporal]]),
                Option('--damos_quota_fail_charge_ratio', nr_args=2,
                       repeatable=True),
                Option('--damos_filter', -1, True,
                       damos_filter_positional_candids),
                ])

def start_candidates(words, cword):
    return damon_param_candidates(words[2:], cword - 2)

def tune_candidates(words, cword):
    return damon_param_candidates(words[2:], cword - 2)

def record_candidates(words, cword):
    return get_candidates(
            words[2:], cword - 2,
            [Option('--out', 1, False, None),
             Option('--snapshot', 2, False, None),
             Option('--timeout', 1, False, None),
             Option('--snapshot_damos_filter', -1, False,
                    damos_filter_positional_candids),
             Option('--do_record', nr_args=-1, repeatable=False,
                    positional_candidates=None,
                    non_positional_candidates=[
                        'access', 'cpu_profile', 'mem_footprint', 'vmas',
                        'proc_stats', 'damon_trace_events']),
             ])

def report_access_candidates(words, cword):
    if cword < 3:
        return []
    candidates = get_candidates(
            words[3:], cword - 3, [
                Option('--input', 1, True),
                Option('--snapshot_damos_filter', -1, True,
                       damos_filter_positional_candids),
                Option('--style', 1, False,
                       [['detailed', 'simple-boxes', 'temperature-sz-hist',
                         'recency-sz-hist', 'cold-memory-tail',
                         'recency-percentiles', 'idle-time-percentiles',
                         'temperature-percentiles', 'cold', 'hot']]
                       ),
                ])
    if candidates:
        return candidates
    # cword is larger than 3.
    prev = words[cword - 1]
    if prev == '--input':
        candidates = ['tried_regions_of', './', '../']
        for f in os.listdir('./'):
            candidates.append('./%s' % f)
        return candidates
    return []

def report_damon_candidates(words, cword):
    return get_candidates(
            words[3:], cword - 3,
            [Option('--format', 1, False, [['json', 'yaml', 'report']]),
             Option('--show_cpu_usage', 0, False, None),
             Option('--kdamonds_summary', 0, False, None),
             ])

def report_sysinfo_candidates(words, cword):
    return get_candidates(
            words[3:], cword - 3,
            [Option('--print', -1, False, None,
                    non_positional_candidates= [
                        'versions', 'fs_info', 'trace_cmd_info', 'perf_info',
                        'interfaces', 'sysfs_features', 'debugfs_features',
                        'trace_features', 'stat_features', 'lru_sort_features',
                        'unavailable_features',
                        'all']
                    ),
             Option('--invalidate_cache', 0, False, None),
             ])

def report_record_info_candidates(words, cword):
    return get_candidates(
            words[3:], cword - 3,
            [Option(name='--raw_numbers', nr_args=0, repeatable=False,
                    positional_candidates=None,
                    non_positional_candidates=None),
             ])

def report_pa_layout_candidates(words, cword):
    return get_candidates(
            words[3:], cword - 3,
            [
                Option(name='--raw_number', nr_args=0, repeatable=False,
                       positional_candidates=None,
                       non_positional_candidates=None),
                Option(name='--numa_node', nr_args=1, repeatable=False,
                       positional_candidates=None,
                       non_positional_candidates=None),
                Option(name='--numa_addr', nr_args=1, repeatable=False,
                       positional_candidates=None,
                       non_positional_candidates=None),
                ])

def report_trace_candidates(words, cword):
    return get_candidates(
            words[3:], cword - 3,
            [
                Option(name='--event', nr_args=-1, repeatable=False,
                       non_positional_candidates=[
                           'damon:damon_aggregated',
                           'damon:damos_before_apply',
                           'damon:damon_monitor_intervals_tune',
                           'damon:damos_esz',
                           'damon:damos_stat_after_apply_interval']),
                ])

def report_candidates(words, cword):
    if cword == 2:
        return ['access', 'damon holistic', 'heatmap', 'sysinfo',
                'record_info', 'pa_layout', 'trace']
    report_type = words[2]
    if report_type == 'access':
        return report_access_candidates(words, cword)
    if report_type == 'damon':
        return report_damon_candidates(words, cword)
    if report_type == 'sysinfo':
        return report_sysinfo_candidates(words, cword)
    if report_type == 'record_info':
        return report_record_info_candidates(words, cword)
    if report_type == 'pa_layout':
        return report_pa_layout_candidates(words, cword)
    if report_type == 'trace':
        return report_trace_candidates(words, cword)
    return []

def monitor_candidates(words, cword):
    candidates = get_candidates(
            words[2:], cword - 2, [
                Option('--report_type', 1, False,
                       [['heats', 'wss', 'holistic']]),
                Option('--delay', 1, False),
                Option('--count', 1, False),
                ])
    return candidates

def help_candidtes(words, cword):
    if cword == 2:
        return ['damon_param_options', 'access_filter_options',
                'access_format_options']
    if cword == 3:
        topic = words[2]
        if topic == 'damon_param_options':
            return ['all', 'monitoring', 'damos']
        if topic == 'access_format_options':
            return ['options', 'record_format_keywords',
                    'snapshot_format_keywords', 'region_format_keywords']
    return []

def print_suggestions(words, cword):
    candidates = []
    if cword == 0:
        return
    if cword == 1:
        candidates = ['start', 'stop', 'tune', 'record', 'report', 'monitor',
                      'help', 'version']
    cmd = words[1]
    if cmd == 'start':
        candidates = start_candidates(words, cword)
    elif cmd == 'tune':
        candidates = tune_candidates(words, cword)
    elif cmd == 'record':
        candidates = record_candidates(words, cword)
    elif cmd == 'report':
        candidates = report_candidates(words, cword)
    elif cmd == 'monitor':
        candidates = monitor_candidates(words, cword)
    elif cmd == 'help':
        candidates = help_candidtes(words, cword)
    if words[cword - 1] != '--help':
        candidates.append('--help')
    print(' '.join(candidates))

def handle_cli_complete():
    '''
    Print command line auto-completion suggestions.  Read
    scripts/damo-completion.sh and src/damo.py to see how this is called.

    Note that this is not supporting full options.  Only commands and options
    that expected to be frequently used are supported.
    '''
    if len(sys.argv) < 4:
        return False
    if sys.argv[1] != '--cli_complete':
        return False
    cword = int(sys.argv[2])
    words = sys.argv[3:]
    print_suggestions(words, cword)
    return True

def main(args):
    fd, tmp_path = tempfile.mkstemp(prefix='damo_completion_script-')
    with open(tmp_path, 'w') as f:
        f.write(completion_script)
    print('It is nearly done, and need one more step.')
    print('Please execute "source %s" on your shell.' % tmp_path)
    print()
    print('Note that this works only for the current shell session.')
    print('For permanent completion setup, save "%s" somewhere,' % tmp_path)
    print('and modify your .bashrc like things to execute the ')
    print('"source" of the saved file for every session.')

def set_argparser(parser):
    return



# SPDX-License-Identifier: GPL-2.0

"Record and report data access pattern in realtime"

import os
import signal
import subprocess
import sys

import _damon
import _damon_args

def cleanup(exit_code=0):
    if target_type == _damon_args.target_type_cmd and cmd_pipe.poll() == None:
        cmd_pipe.kill()
    if target_type != 'ongoing':
        damo = sys.argv[0]
        subprocess.call(
                [damo, 'stop'],
                # DAMON may already stopped, but that's fine.  Ignore error
                # message.
                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    exit(exit_code)

def sighandler(signum, frame):
    print('\nsignal %s received' % signum)
    cleanup()

def main(args):
    _damon.ensure_root_permission()

    global target_type
    global cmd_pipe

    signal.signal(signal.SIGINT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)

    damo = sys.argv[0]
    target = args.target
    if target != 'ongoing':
        target_type = _damon_args.deduced_target_type(target)
        if target_type is None:
            print('invalid target \'%s\'' % target)
            exit(1)
        if target_type == _damon_args.target_type_explicit and \
                target == 'paddr':
            pass
        elif target_type == _damon_args.target_type_cmd:
            cmd_pipe = subprocess.Popen(
                    target, shell=True, executable='/bin/bash',
                    stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
            target = cmd_pipe.pid

        if subprocess.call([damo, 'start', '%s' % target]) != 0:
            print('starting DAMON fail')
            exit(1)
    else:
        target_type = 'ongoing'

    record_cmd = [damo, 'record', '--timeout', '%f' % args.delay, 'ongoing']

    report_cmd = [damo, 'report']
    if args.report_type == 'heats':
        report_cmd += ['heatmap', '--resol', '10', '80',
                       '--time_range', '%f s' % (args.delay * -1), '0 s',
                       'guided_end']
    else:
        report_cmd += [args.report_type]
        if args.report_type == 'holistic':
            report_cmd += ['--heatmap_time_last_n_sec', '%f' % args.delay]

    nr_reports = 0
    while not args.count or nr_reports < args.count:
        if (target_type == _damon_args.target_type_cmd and
                cmd_pipe.poll() != None):
            break
        try:
            subprocess.check_output(record_cmd, stderr=subprocess.STDOUT)
        except Exception as e:
            print('Recording fail (%s)' % e)
            if target_type == 'ongoing':
                print('Maybe you stopped DAMON?')
            break
        try:
            output = subprocess.check_output(report_cmd).decode()
            if args.report_type == 'heats':
                for line in output.strip().split('\n'):
                    if not line.startswith('#'):
                        print(line)
            else:
                print(output)
        except Exception as e:
            print('Report generating fail (%s)' % e)
            if target_type == 'ongoing':
                print('Maybe you stopped DAMON?')
            break
        nr_reports += 1

    cleanup()

def set_argparser(parser):
    parser.add_argument('target', type=str, metavar='<target>',
            help='monitoring target (command, pid, \'paddr\' or \'ongoing\')')
    parser.add_argument(
            '--report_type', type=str, choices=['heats', 'wss', 'holistic'],
            default='heats', help='report type')
    parser.add_argument('--delay', type=float, metavar='<seconds>', default=3,
            help='deplay between updates in seconds.')
    parser.add_argument('--count', type=int, metavar='<count>', default=0,
            help='number of updates.')



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0

import collections
import copy
import json
import unittest

import _test_damo_common

_test_damo_common.add_damo_dir_to_syspath()

import _damon

class TestDamon(unittest.TestCase):
    def test_kvpairs_transition(self):
        target = _damon.DamonTarget(1234, [_damon.DamonRegion(10, 20)])
        target_kvpairs = target.to_kvpairs()
        self.assertEqual(type(target_kvpairs), collections.OrderedDict)
        self.assertEqual(list(target_kvpairs.keys()),
                ['pid', 'obsolete', 'regions'])
        self.assertEqual(target,
                _damon.DamonTarget.from_kvpairs(target_kvpairs))

        damos = _damon.Damos(
                _damon.DamosAccessPattern([0, 10], [5, 8], _damon.unit_percent,
                    [54, 88], _damon.unit_usec),
                'pageout', None,
                None,
                _damon.DamosQuotas(100, 1024, 1000, [80, 76, 24]),
                _damon.DamosWatermarks('free_mem_rate', 5000000, 800, 500,
                    200),
                [_damon.DamosFilter('memcg', True, memcg_path='/foo/bar'),
                    _damon.DamosFilter('anon', False)],
                    None, None)
        damos_kvpairs = damos.to_kvpairs()
        self.assertEqual(type(damos_kvpairs), collections.OrderedDict)
        self.assertEqual(list(damos_kvpairs.keys()),
                ['action', 'dests', 'access_pattern', 'apply_interval_us',
                 'quotas', 'watermarks', 'filters', 'stats'])
        self.assertEqual(list(damos_kvpairs['stats'].keys()),
                ['nr_tried', 'sz_tried', 'nr_applied', 'sz_applied',
                 'sz_ops_filter_passed', 'qt_exceeds', 'nr_snapshots',
                 'max_nr_snapshots'])
        self.assertEqual(damos, _damon.Damos.from_kvpairs(damos_kvpairs))

        ctx = _damon.DamonCtx('paddr', [target],
                _damon.DamonIntervals(5000, 100000, 1000000),
                _damon.DamonNrRegionsRange(10, 1000),
                [damos])
        ctx_kvpairs = ctx.to_kvpairs()
        self.assertEqual(type(ctx_kvpairs), collections.OrderedDict)
        self.assertEqual(list(ctx_kvpairs.keys()),
                ['ops', 'addr_unit', 'targets', 'intervals', 'nr_regions',
                 'probes', 'sample_control', 'pause', 'schemes'])
        self.assertEqual(ctx, _damon.DamonCtx.from_kvpairs(ctx_kvpairs))

        kdamond = _damon.Kdamond('off', 123, [ctx])
        kdamond_kvpairs = kdamond.to_kvpairs()
        self.assertEqual(type(kdamond_kvpairs), collections.OrderedDict)
        self.assertEqual(list(kdamond_kvpairs.keys()),
                ['state', 'pid', 'refresh_ms', 'contexts'])
        self.assertEqual(kdamond,
            _damon.Kdamond.from_kvpairs(kdamond_kvpairs))

    def test_damos_default_immutable(self):
        damos = _damon.Damos()
        before = damos.quotas.time_ms
        damos.quotas.time_ms = 123
        damos = _damon.Damos()
        self.assertEqual(before, damos.quotas.time_ms)

    def test_damos_eq(self):
        damos = _damon.Damos(
                access_pattern=_damon.DamosAccessPattern([4096,
                    18446744073709551615], [0.0, 0.0], _damon.unit_percent,
                    [1000000, 4294900000], _damon.unit_usec),
                action='stat', target_nid=None,
                apply_interval_us=None,
                quotas=_damon.DamosQuotas(time_ms=0, sz_bytes=584792941,
                    reset_interval_ms=1000, weights=[0,0,0]),
                watermarks=_damon.DamosWatermarks(
                    _damon.damos_wmarks_metric_none,0,0,0,0),
                filters=[], stats=None)
        damos2 = copy.deepcopy(damos)
        self.assertEqual(damos, damos2)

        damos2.filters = [_damon.DamosFilter(filter_type='memcg',
            matching=True, memcg_path='/foo/bar/')]
        self.assertNotEqual(damos, damos2)

        intervals = _damon.DamonIntervals(5000, 100000, 1000000)
        pattern_human = _damon.DamosAccessPattern([123, 456],
                [15, 35], _damon.unit_percent,
                [5000000, 19000000], _damon.unit_usec)
        pattern_machine = _damon.DamosAccessPattern([123, 456],
                [3, 7], _damon.unit_samples,
                [50, 190], _damon.unit_aggr_intervals)

        damos.access_pattern = pattern_human
        damos2 = copy.deepcopy(damos)
        damos2.access_pattern = pattern_machine
        self.assertNotEqual(damos, damos2)
        self.assertTrue(damos.effectively_equal(damos2, intervals))
        self.assertTrue(damos.effectively_equal(damos2, intervals))

    def test_damos_action_validity(self):
        exception_raised = False
        try:
            _damon.Damos(action='foo')
        except:
            exception_raised = True
        self.assertTrue(exception_raised)
        for action in _damon.damos_actions:
            exception_raised = False
            try:
                _damon.Damos(action=action)
            except:
                exception_raised = True
            self.assertFalse(exception_raised)

    def test_damon_intervals(self):
        self.assertEqual(_damon.DamonIntervals(),
                _damon.DamonIntervals(5000, 100000, 1000000))

        _test_damo_common.test_input_expects(self,
                lambda x: _damon.DamonIntervals.from_kvpairs(json.loads(x)),
                {
                    json.dumps({'sample_us': 5000, 'aggr_us': 100000,
                        'ops_update_us': 1000000}):
                    _damon.DamonIntervals(5000, 100000, 1000000),
                    json.dumps({'sample_us': '5ms', 'aggr_us': '0.1s',
                        'ops_update_us': '1s'}):
                    _damon.DamonIntervals(5000, 100000, 1000000)})

        self.assertEqual('%s' % _damon.DamonIntervals(5000, 100000, 1000000),
        'sample 5 ms, aggr 100 ms, update 1 s')
        self.assertEqual('%s' %
                _damon.DamonIntervals(5000, 100000, 1000000).to_str(True),
                'sample 5000, aggr 100000, update 1000000')

        intervals_kvpairs = _damon.DamonIntervals(5000, 100000, 1000000).to_kvpairs()
        del intervals_kvpairs['intervals_goal']
        self.assertEqual(
                intervals_kvpairs,
                {'sample_us': '5 ms', 'aggr_us': '100 ms',
                 'ops_update_us': '1 s',
                 })

    def test_damon_nr_regions_range(self):
        self.assertEqual(_damon.DamonNrRegionsRange(),
                _damon.DamonNrRegionsRange(10, 1000))

        expect = _damon.DamonNrRegionsRange(10, 1000)
        _test_damo_common.test_input_expects(self,
                lambda x: _damon.DamonNrRegionsRange(*x),
                {
                    tuple([10, '1000']): expect,
                    tuple(['10', '1000']): expect,
                    tuple(['10', '1,000']): expect})

        _test_damo_common.test_input_expects(self,
                lambda x: _damon.DamonNrRegionsRange.from_kvpairs(
                    json.loads(x)),
                {
                    json.dumps({'min': 10, 'max': 1000}):
                    expect,
                    json.dumps(
                        {'min': '10', 'max': '1000'}):
                    expect,
                    json.dumps(
                        {'min': '10', 'max': '1,000'}):
                    expect})

        self.assertEqual('%s' % _damon.DamonNrRegionsRange(10, 1000),
                '[10, 1,000]')
        self.assertEqual(
                '%s' % _damon.DamonNrRegionsRange(10, 1000).to_str(True),
                '[10, 1000]')

        self.assertEqual(_damon.DamonNrRegionsRange(10, 1000).to_kvpairs(),
                {'min': '10', 'max': '1,000'})
        self.assertEqual(_damon.DamonNrRegionsRange(10, 1000).to_kvpairs(
                        raw=True),
                {'min': '10', 'max': '1000'})

    def test_damon_region(self):
        _test_damo_common.test_input_expects(self,
                lambda x: _damon.DamonRegion(*x),
                {
                    tuple([123, '456']): _damon.DamonRegion(123, 456),
                    tuple(['123', '456']): _damon.DamonRegion(123, 456),
                    tuple(['1,234', '4,567']): _damon.DamonRegion(1234, 4567),
                    tuple(['1K', '4K']): _damon.DamonRegion(1024, 4096),
                    })

        _test_damo_common.test_input_expects(self,
                lambda x: _damon.DamonRegion.from_kvpairs(json.loads(x)),
                {
                    json.dumps({'start': '123', 'end': '456'}):
                    _damon.DamonRegion(123, 456),
                    json.dumps({'start': '1234', 'end': '4567'}):
                    _damon.DamonRegion(1234, 4567),
                    json.dumps({'start': '1,234', 'end': '4,567'}):
                    _damon.DamonRegion(1234, 4567),
                    json.dumps({'start': '1K', 'end': '4K'}):
                    _damon.DamonRegion(1024, 4096),
                    })

        self.assertEqual('%s' % _damon.DamonRegion(123, 456),
                '[123, 456) (333 B)')
        self.assertEqual('%s' % _damon.DamonRegion(123, 456).to_str(raw=True),
                '[123, 456) (333)')

        self.assertEqual(_damon.DamonRegion(1234, 5678).to_kvpairs(),
                {'start': '1,234', 'end': '5,678'})
        self.assertEqual(_damon.DamonRegion(1234, 5678).to_kvpairs(raw=True),
                {'start': '1234', 'end': '5678'})

    def test_damos_access_pattern(self):
        self.assertEqual(_damon.DamosAccessPattern(),
                _damon.DamosAccessPattern(['min', 'max'],
                    ['min', 'max'], _damon.unit_percent,
                    ['min', 'max'], _damon.unit_usec))

        intervals = _damon.DamonIntervals(5000, 100000, 1000000)
        pattern_human = _damon.DamosAccessPattern([123, '4,567'],
                [15, '10,000'], _damon.unit_percent,
                ['5,000,000', '190,000,000'], _damon.unit_usec)
        pattern_machine = _damon.DamosAccessPattern([123, 4567],
                [3, '2,000'], _damon.unit_samples,
                [50, '1,900'], _damon.unit_aggr_intervals)

        self.assertEqual(
                pattern_human.converted_for_units(
                    _damon.unit_samples,
                    _damon.unit_aggr_intervals, intervals),
                pattern_machine)
        self.assertEqual(
                pattern_machine.converted_for_units(
                    _damon.unit_percent, _damon.unit_usec, intervals),
                pattern_human)

        self.assertTrue(
                pattern_human.effectively_equal(pattern_machine, intervals))

        for converted in [
                pattern_human.converted_for_units(_damon.unit_samples,
                    _damon.unit_aggr_intervals, intervals),
                pattern_machine.converted_for_units(
                    _damon.unit_percent, _damon.unit_usec, intervals)]:
            self.assertEqual(type(converted.nr_acc_min_max[0].samples), int)
            self.assertEqual(type(converted.nr_acc_min_max[1].samples), int)
            self.assertEqual(type(converted.nr_acc_min_max[0].percent), int)
            self.assertEqual(type(converted.nr_acc_min_max[1].percent), int)
            self.assertEqual(type(converted.age_min_max[0].aggr_intervals), int)
            self.assertEqual(type(converted.age_min_max[1].aggr_intervals), int)
            self.assertEqual(type(converted.age_min_max[0].usec), int)
            self.assertEqual(type(converted.age_min_max[1].usec), int)

        exception_raised = False
        try:
            _damon.DamosAccessPattern(nr_accesses=[0, 100], nr_accesses_unit='foo')
        except:
            exception_raised = True
        self.assertTrue(exception_raised)
        exception_raised = False
        try:
            _damon.DamosAccessPattern(age=[0, 100], age_unit='bar')
        except:
            exception_raised = True
        self.assertTrue(exception_raised)

    def test_damos_quotas(self):
        self.assertEqual(_damon.DamosQuotas(),
                _damon.DamosQuotas(0, 0, 'max', [0, 0, 0]))

    def test_damos_watermarks(self):
        self.assertEqual(_damon.DamosWatermarks(),
                _damon.DamosWatermarks(_damon.damos_wmarks_metric_none,
                    0, 0, 0, 0))

    def test_damon_nr_accesses(self):
        intervals = _damon.DamonIntervals('5ms', '100ms', '1s')

        nr_acc = _damon.DamonNrAccesses(4, _damon.unit_samples)
        nr_acc_2 = _damon.DamonNrAccesses(4, _damon.unit_samples)
        self.assertEqual(nr_acc, nr_acc_2)
        nr_acc_2 = _damon.DamonNrAccesses(6, _damon.unit_samples)
        self.assertNotEqual(nr_acc, nr_acc_2)
        nr_acc.add_unset_unit(intervals)
        self.assertEqual(nr_acc.samples, 4)
        self.assertEqual(nr_acc.percent, 20)
        self.assertEqual('%s' % nr_acc.to_str(_damon.unit_samples, raw=False),
                '4 %s' % _damon.unit_samples)
        self.assertEqual('%s' % nr_acc.to_str(_damon.unit_percent, raw=False),
                '20 %')

        nr_acc = _damon.DamonNrAccesses(20, _damon.unit_percent)
        nr_acc_2 = _damon.DamonNrAccesses(20, _damon.unit_percent)
        self.assertEqual(nr_acc, nr_acc_2)
        nr_acc_2 = _damon.DamonNrAccesses(25, _damon.unit_percent)
        self.assertNotEqual(nr_acc, nr_acc_2)

        nr_acc.add_unset_unit(intervals)
        self.assertEqual(nr_acc.samples, 4)
        self.assertEqual(nr_acc.percent, 20)
        self.assertEqual('%s' % nr_acc.to_str(_damon.unit_samples, raw=False),
                '4 %s' % _damon.unit_samples)
        self.assertEqual('%s' % nr_acc.to_str(_damon.unit_percent, raw=False),
                '20 %')

        nr_acc = _damon.DamonNrAccesses.from_kvpairs(
                {'samples': 12, 'percent': 45})
        self.assertEqual(nr_acc.samples, 12)
        self.assertEqual(nr_acc.percent, 45)

        nr_acc = _damon.DamonNrAccesses.from_kvpairs(
                {'samples': '12', 'percent': '45%'})
        self.assertEqual(nr_acc.samples, 12)
        self.assertEqual(nr_acc.percent, 45)

    def test_damon_age(self):
        intervals = _damon.DamonIntervals('5ms', '100ms', '1s')

        age = _damon.DamonAge(15, _damon.unit_aggr_intervals)
        age_2 = _damon.DamonAge(15, _damon.unit_aggr_intervals)
        self.assertEqual(age, age_2)
        age_2 = _damon.DamonAge(18, _damon.unit_aggr_intervals)
        self.assertNotEqual(age, age_2)
        age.add_unset_unit(intervals)
        self.assertEqual(age.aggr_intervals, 15)
        self.assertEqual(age.usec, 1500000)
        self.assertEqual(
                '%s' % age.to_str(_damon.unit_aggr_intervals, raw=False),
                '15 %s' % _damon.unit_aggr_intervals)
        self.assertEqual('%s' % age.to_str(_damon.unit_usec, raw=False),
                '1 s 500 ms')

        age = _damon.DamonAge.from_kvpairs({'usec': 12, 'aggr_intervals': 456})
        self.assertEqual(age.usec, 12)
        self.assertEqual(age.aggr_intervals, 456)

    def test_best_effort_target_arrange(self):
        updated_targets = [
                _damon.DamonTarget(pid=1),
                _damon.DamonTarget(pid=2, obsolete=True),
                _damon.DamonTarget(pid=3),
                _damon.DamonTarget(pid=4, obsolete=True),
                ]
        new_targets = [
                _damon.DamonTarget(pid=5),
                _damon.DamonTarget(pid=6),
                _damon.DamonTarget(pid=7),
                ]
        expect_targets = [
                _damon.DamonTarget(pid=1),
                _damon.DamonTarget(pid=5),
                _damon.DamonTarget(pid=3),
                _damon.DamonTarget(pid=6),
                _damon.DamonTarget(pid=7),
                ]
        result_targets = _damon.best_effort_target_arrange(
                updated_targets, new_targets)
        self.assertEqual(result_targets, expect_targets)

if __name__ == '__main__':
    unittest.main()
